#!/usr/bin/python3
import os
import math
import argparse

from pathlib import Path

from pykinetic.classes import Energy,SimulationParameters,ConvergenceParameters
from pykinetic.writers import CplusplusWriter
import pykinetic.writers as writers
from pykinetic.utils import (BiasedChemicalSystem, write_indexfile, 
                              calc_standard_state_correction)
from pykinetic.userinput import populate_chemicalsystem_fromfiles

__version__ = "0.1.0"

WRITERS = {'python':writers.python.Batch,
           'c++':CplusplusWriter}

def create_parser():
    """ Create a Command line argument parser """
    description =""" Creates a python script that contains a
    system of differential equations that represent a Chemical System
    and the tools to solve it numerically"""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("compounds",type=Path,
                        help="File with the compounds and Energies")
    parser.add_argument("reactions",type=Path, 
                        help="File with the reactions, energies and/or TSs")
    parser.add_argument("outfile", type=Path, help="Output script")
    parser.add_argument("--relative",
                        action='store_true',
                        default=False,
                        help="""Whether the energy in the reactions file is 
                        relative to reactants. Otherwise the absolute value of 
                        the TS structure is assumed to correspond to the 
                        provided energy""")
    parser.add_argument("-I","--IndexFile",
                        action="store_true",
                        default=False,
                        help="""If Enabled creates a txt file named 
                        'OutFile'.index with the indices used for reactions,
                        compounds and TSs""",
                        )
    parser.add_argument("-Ir","--indexfile-relative",
                        action="store_true",
                        dest="indexfile_relative",
                        default=False,
                        help="""Enforces relative barriers in the specified 
                        default energy unit in the indexfile.""")
    parser.add_argument("-T","--Temperature", 
                        nargs=2,
                        metavar=("Temp","Unit"),
                        default=(25.0,'C'),
                        help="Temperature and unit ('C'=Celsius and 'K'=Kelvin)")
    parser.add_argument("--unit",
                        default='kcal/mol',
                        help="""Energy unit used in all the energy values whose 
                        unit is not specified""", 
                        choices=Energy._units)
    parser.add_argument("--bias",
                        nargs='?',
                        default=0.0,
                        const='Standard State',
                        help="""Applies a bias to all energies. If no argument 
                        is provided it applies Standard State correction. 
                        Otherwise, uses the specified value """)
    parser.add_argument("--convergence", 
                        type=Path,
                        default=None,
                        help="Path to the file with the Convergence parameters")
    parser.add_argument("--simulation",
                        type=Path,
                        default=None,
                        help="Path to the file with the Simulation parameters")
    parser.add_argument("--tail",
                        type=Path,
                        default=None,
                        help="""Path to the non-standard Template for the ending
                         part of the generated script""")
    parser.add_argument("--header",
                        type=Path,
                        default=None,
                        help=""" Path to the non-standard Template for the head 
                        of the generated script""")
    parser.add_argument("--writer",
                        choices=['python','c++'],
                        default='python',
                        help="format for the output model")
    
    return parser
def parse_arguments(parser):
    """ Check the arguments of the parser and prepare them for their use """
    args = parser.parse_args()

    # Read the header if it has been specified
    if args.header is not None: 
        with open(args.header) as F:
            txt = F.read()
        args.header = txt
    # Read the tail if it has been specified
    if args.tail is not None: 
        with open(args.tail) as F:
            txt = F.read()
        args.tail = txt

    # Handle Temperature
    if args.Temperature[1] == 'C':
        T = float(args.Temperature[0]) + 273.15
    elif args.Temperature[1] == 'K':
        T = float(args.Temperature[0])
    else:
        msg = "Temperature in {} not implemented".format(args.Temperature[1])
        raise NotImplementedError(msg)
    args.Temperature = T

    # Handle bias
    if args.bias == 'Standard State':
        bias = calc_standard_state_correction(T)
        bias.to_unit(args.unit)
    else:
        bias = Energy(args.bias,args.unit)
    args.bias = bias

    # Handle simulation parameters
    if args.simulation is None: 
        sim_params = SimulationParameters()
    else:
        sim_params = SimulationParameters.read_from(args.simulation)
    args.simulation = sim_params

    # Handle convergence parameters
    if args.convergence is None:
        conv_params = ConvergenceParameters()
    else:
        conv_params = ConvergenceParameters.read_from(args.convergence)
    args.convergence = conv_params

    if not args.IndexFile and args.indexfile_relative: 
        args.IndexFile = True

    return args


def main():
    parser = create_parser()
    args = parse_arguments(parser)
    unit = args.unit
    T = args.Temperature
    # Initialize the Writer
    writer_cls = WRITERS[args.writer]
    writer = writer_cls(conc_var='x',mb_var='dxdt',fun_var='model',
                        jac_var='Jac',jac_fun_var='jacobian',
                        header=args.header,tail=args.tail)
    writer.set_parameters(simulation=args.simulation,
                          convergence=args.convergence)
    # Initialize the ChemicalSystem 
    chemsys = BiasedChemicalSystem(bias=args.bias,T=args.Temperature,unit=unit)
    populate_chemicalsystem_fromfiles(chemsys,
                                      file_c=args.compounds,
                                      file_r=args.reactions,
                                      energy_unit=unit,
                                      relativeE=args.relative)
    chemsys.apply_bias()

    data_file_name = f'{args.outfile.stem}.data' 
    writer.parameters['out_filename'] = data_file_name
    writer.write(chemsys,args.outfile)

    if args.IndexFile:
        stem = args.outfile.stem
        index_file = args.outfile.parent / f'{stem}.index'
        isrelative = args.indexfile_relative or args.relative
        write_indexfile(chemsys,index_file, isrelative=isrelative)

if __name__ == '__main__':
    main()



from pykinetic.classes import Compound, Energy, ChemicalSystem, Reaction, TransitionState
from pykinetic.userinput import *
import unittest
from unittest.mock import mock_open, patch, MagicMock

class InputTest(unittest.TestCase):
    def setUp(self):
        TransitionState.defaultname = 0
        DiffusionTS.defaultname = 0
    def test_populate_chemicalsystem_fromfiles(self):
        module_name = 'pykinetic.userinput'
        default_unit = 'kcal/mol'
        # Prepare the outputs of the mocks
        compounds = [['A','0.0',default_unit],
                     ['B','0.0',default_unit],
                     ['C','1.0'],
                     ['D','-2.0',default_unit]]
        reactions1 = [[(['A','B'],'<=>',['C',],f'10.0 {default_unit}'), 
                      (['C',],'=>',['D',],'TSa')],
                     ['TSa     20',]]
        reactions2 = [[(['A','B'],'<=>',['C',],f'10.0 {default_unit}'), 
                      (['C',],'=>',['D',],'TSa')],
                     ['TSa     18',]]
        # Prepare the solution
        chemsys = ChemicalSystem()
        A = Compound('A',Energy('0.0',default_unit))
        B = Compound('B',Energy('0.0',default_unit))
        C = Compound('C',Energy('1.0',default_unit))
        D = Compound('D',Energy('-2.0',default_unit))
        TS00 = TransitionState(Energy('10.0',default_unit),label='TS00')
        TS01 = TransitionState(Energy('10.0',default_unit),label='TS01')
        TSa = TransitionState(Energy('20.0',default_unit),label='TSa')
        r1 = Reaction((A,B),(C,),TS=TS00)
        r2 = Reaction((C,),(A,B),TS=TS00)
        r3 = Reaction((C,),(D,),TS=TSa)
        for compound in [A,B,C,D]: 
            chemsys.cadd(compound)
        for reaction in [r1,r2,r3]: 
            chemsys.radd(reaction,False)
        chemsys.rupdate()
        chemicalsystem1 = ChemicalSystem()
        chemicalsystem2 = ChemicalSystem()
        # run the function with the mocks
        with patch(f'{module_name}.read_compounds',return_value=compounds):
            with patch(f'{module_name}.read_reactions',return_value=reactions1):
                populate_chemicalsystem_fromfiles(chemicalsystem1,
                                                  'file_compounds',
                                                  'file_reactions',
                                                  default_unit,
                                                  False)
            with patch(f'{module_name}.read_reactions',return_value=reactions2):
                populate_chemicalsystem_fromfiles(chemicalsystem2,
                                                  'file_compounds',
                                                  'file_reactions',
                                                  default_unit,
                                                  False)
        attributes =['compounds','reactions','Name2Compound','transitionstates']
        for attr in attributes: 
            with self.subTest(energy_parse='absolute',attribute=attr):
                test = getattr(chemicalsystem1,attr)
                sol = getattr(chemsys,attr)
                self.assertEqual(test, sol)
        _ = attributes.pop(-1)
        for attr in attributes: 
            with self.subTest(energy_parse='relative',attribute=attr):
                test = getattr(chemicalsystem2,attr)
                sol = getattr(chemsys,attr)
                self.assertEqual(test, sol)
        with self.subTest(energy_parse='relative',attribute='transitionstates'):
            test = chemicalsystem2.transitionstates
            sol = [TS01,TSa]
            self.assertEqual(test,sol)
        for compound in chemicalsystem1.compounds:
            compound2 = chemicalsystem2.Name2Compound[compound.label] 
            with self.subTest(test='energy consistency',compound=compound.label):
                self.assertEqual(compound.energy,compound2.energy)
        ts = chemicalsystem1.Name2TS['TS00']
        ts2 = chemicalsystem2.Name2TS['TS01']
        with self.subTest(test='energy consistency',TS='defaultname'):
            self.assertTrue(ts.energy,ts2.energy)
        with self.subTest(test='energy consistency',TS='TSa'):
            ts = chemicalsystem1.Name2TS['TSa']
            ts2 = chemicalsystem2.Name2TS['TSa']
            self.assertTrue(ts.energy,ts2.energy)
    def test_read_compounds(self):
        module_name = 'pykinetic.userinput'
        raw_compounds = """
        A   20      kcal/mol    scan
        B   10      hartree
        C   1   
        D   0.5     eV
        """
        solutions = [['A','20','kcal/mol','scan'],
                     ['B','10','hartree'],
                     ['C','1'],
                     ['D','0.5','eV']]
        with patch(f'{module_name}.open', mock_open(read_data=raw_compounds), 
                    create=True) as m:
            tests = read_compounds('mock_file')
        for i,(test,solution) in enumerate(zip(tests,solutions)):
            with self.subTest(compound='reactions'):
                self.assertEqual(test,solution)
    def test_create_compounds(self):
        default_unit = 'kcal/mol'
        compounds = [['A','20','kcal/mol','scan'],
                     ['B','10','hartree'],
                     ['C','1'],
                     ['D','0.5','eV']]
        solutions = [Compound('A',Energy('20','kcal/mol'),scannable=True),
                     Compound('B',Energy('10','hartree')),
                     Compound('C',Energy('1',default_unit)),
                     Compound('D',Energy('0.5','eV'))]
        tests = create_compounds(compounds,default_unit)
        for test,solution in zip(tests,solutions):
            with self.subTest(compound=solution.label): 
                self.assertEqual(test,solution)
                self.assertEqual(test.energy,solution.energy)
                self.assertEqual(test.scannable,solution.scannable)
        with self.subTest(test='raises error'):
            with self.assertRaises(ValueError):
                _ = create_compounds(compounds + compounds)
    def test_read_reactions(self):
        module_name = 'pykinetic.userinput'
        reactions = """
                    A  +  B    <=>   C    +   D    !25.0 kcal/mol 
                    A  +  B    <=>   C    +   D    !25.0 
                    A  +  B    <=>   C    +   D    !TS1 
                    A          <=>   C             !TS1 
                    TS1     25
                    TS2     72      J/mol
                    TS3     1.2     eV      scan
                    """
        solutions = [[(['A','B'],'<=>',['C','D'],'25.0 kcal/mol'), 
                      (['A','B'],'<=>',['C','D'],'25.0'),
                      (['A','B'],'<=>',['C','D'],'TS1'), 
                      (['A',   ],'<=>',['C',   ],'TS1')],
                      ['TS1     25',
                       'TS2     72      J/mol',
                       'TS3     1.2     eV      scan']]
        with patch(f'{module_name}.open', mock_open(read_data=reactions), 
                    create=True) as m:
            test = read_reactions('mock_file')
        with self.subTest(block='reactions'):
            self.assertEqual(test[0],solutions[0])
        with self.subTest(block='TSs'):
            self.assertEqual(test[1],solutions[1])
    def test_create_TS_dict(self):
        default_unit = 'kcal/mol'
        lines = ['TS1    -22 ',
                 'TS2    1.2    eV',
                 'TS3    25.0',
                 'TS4    -10.0  scan',
                 'TS5    23.0   kcal/mol',
                 'TS6    0.01   hartree',
                 'TS7    2  kcal/mol    scan']
        solution = {'TS1':(Energy(-22,'kcal/mol'),False),
                    'TS2':(Energy(1.2,'eV'),False),
                    'TS3':(Energy(25.0,'kcal/mol'),False),
                    'TS4':(Energy(-10.0,'kcal/mol'),True),
                    'TS5':(Energy(23.0,'kcal/mol'),False),
                    'TS6':(Energy(0.01,'hartree'),False),
                    'TS7':(Energy(2,'kcal/mol'),True)}
        TS_dict = create_TS_dict(lines,default_unit)
        for key,item in TS_dict.items():
            with self.subTest(key=key):
                self.assertEqual(item,solution.get(key,None))
    def test_prepare_inline_TS(self): 
        default_unit = 'kcal/mol'
        raw_data = ['TS1',
                    'TS2',
                    '25.0',
                    '-10.0',
                    '23.0 kcal/mol',
                    '0.01 hartree',
                    '2 kcal/mol scan']
        TSdict = {'TS1':(Energy(10.0,'kcal/mol'),True),
                  'TS2':(Energy(1.0,'eV'),False)}
        solutions = [('TS1',TSdict['TS1'][0],TSdict['TS1'][1]),
                     ('TS2',TSdict['TS2'][0],TSdict['TS2'][1]),
                     (None,Energy(25.0,default_unit),False),
                     (None,Energy(-10.0,default_unit),False),
                     (None,Energy(23.0,'kcal/mol'),False),
                     (None,Energy(0.01,'hartree'),False),
                     (None,Energy(2,'kcal/mol'),True)]
        for arg,sol in zip(raw_data,solutions):
            with self.subTest(arg=arg): 
                test = prepare_inline_TS(arg,TSdict,default_unit)
                self.assertEqual(test,sol)
    def test_split_reaction_line(self):
        reactions = """
        A  +  B    <=>   C    +   D    !25.0 kcal/mol 
        A  +  B    <=>   C    +   D    !25.0 
        A  +  B    <=>   C    +   D    !TS1
        A          <=>   C             !TS1
        A  +  B    <=>  [C+]  +   D    !TS1
        A  +  B    =>    C    +   D    !TS1
        C  +  D    <=    A    +   B    !TS1
        A  +  B    <d>   C    +   D    !25.0 
        A        <ERROR>      C        !TSError
        """.strip().split('\n')
        solutions = [(['A','B'],'<=>',['C','D'],'25.0 kcal/mol'),
                     (['A','B'],'<=>',['C','D'],'25.0'),
                     (['A','B'],'<=>',['C','D'],'TS1'),
                     (['A',],'<=>',['C',],'TS1'),
                     (['A','B'],'<=>',['[C+]','D'],'TS1'),
                     (['A','B'],'=>',['C','D'],'TS1'),
                     (['C','D'],'<=',['A','B'],'TS1'),
                     (['A','B'],'<d>',['C','D'],'25.0')]
        for reaction,solution in zip(reactions,solutions): 
            # The last reaction is not iterated here by design
            with self.subTest(reaction=reaction):
                test = split_reaction_line(reaction)
                self.assertEqual(test,solution)
        with self.subTest(test='error'):
            with self.assertRaises(ValueError):
                _ = split_reaction_line(reactions[-1])

    
    


"""
This module contains the Writer classes that are in charge of translating the 
provided chemical system to the different languages or formats as well as the 
Base class to inherit from, 'Writer'. 
Currently only python and c++ are supported. This module depends on the 
template files that come with the libary.
"""

from abc import abstractmethod
from pathlib import Path
from collections import defaultdict
from pkg_resources import resource_filename

TEMPLATES_PATH = Path(resource_filename('pykinetic','templates'))

def Indent(lines,tab='    ',level=1):
    """
    Takes an iterable of strings, each representing a line and returns them with
    the indentation level and character specified.
    """
    Out = []
    for line in lines:
        Out.append(f'{tab*level}{line}')
    return Out

class Writer(object):
    """
    Base class of a writer object. A writer object allows the customized 
    translation of a kinetic model into a script to carry out its simulation in 
    a specific language and under a specific system constraints.  
    """
    def __init__(self, conc_var='x', mb_var='dxdt', fun_var='model',
                 jac_var='Jac',jac_fun_var='jacobian',header=None,tail=None):
        self.x = conc_var
        self.dxdt = mb_var
        self.f = fun_var
        self.jac = jac_var
        self.jac_f = jac_fun_var
        if header is None:
            self._load_default_header()
        elif header:
            self._header = header
        else:
            self._header = ''
        if tail is None:
            self._load_default_tail()
        elif tail:
            self._tail = tail
        else:
            self._tail = ''
        self.clear()
        self.parameters = defaultdict(lambda : 'MISSING')

    # Defaults Load
    @abstractmethod
    def _load_default_tail(self):
        pass
    @abstractmethod
    def _load_default_header(self):
        pass

    # methods for parameter reading
    def set_parameters(self,simulation=None,convergence=None):
        if simulation is not None:
            self.parameters.update(simulation)
        if convergence is not None:
            self.parameters.update(convergence)

    # methods for object -> str transformations
    def constant(self,reaction,value_format):
        """
        Takes in a reaction and returns 2 expressions, the string of the 
        variable and the string with the value of the constant. I.e.

        # Reaction(1 => 2)

        A = <ElementalStep 1 of Type '=>' with reacts=(1,) products(2,)>

        'k01', '1.00000000' = PythonWriter.constant(A)

        Parameters
        ----------
        reaction : Reaction
            Any reaction

        Returns
        -------
        var,expr
            variable and expresion of the reaction
        """
        return f'k{reaction.key:02.0f}', value_format.format(reaction.k)

    @abstractmethod
    def ratelaw(self,reaction):
        """
        Takes in a reaction and returns 2 expressions, the string of the 
        variable and the string of the ratelaw expression. I.e.
        # Reaction(1 => 2)
        A = <ElementalStep 1 of Type '=>' with reacts=(1,) products(2,)>
        'r01', 'k01*x[1]' = PythonWriter.ratelaw(A)

        Parameters
        ----------
        reaction : Reaction
            Any reaction

        Returns
        -------
        var,expr
            variable and expresion of the reaction
        """
        pass
    @abstractmethod
    def ratelaw_partial(self,reaction,compound):
        """
        Takes in a reaction and returns and the string of the partial derivative
        the ratelaw with respect to the concentration of compound. I.e.

        # Reaction(1 => 2), Compound 1

        A = <ElementalStep 1 of Type '=>' with reacts=(1,) products(2,)>

        # r01,'k01*x[1]'ratelaw(A)

        'k01' = PythonWriter.ratelaw_partial(A)

        Parameters
        ----------

        reaction : Reaction
            Any reaction
        compound : Compound
            Any compound

        Returns
        -------
        
        expr
            variable and expresion of the reaction
        """
        pass
    @abstractmethod
    def massbalance(self,Massbalance):
        """
        Takes in a mass balance and returns 2 expressions, the string of the
        variable and the string of the expression. I.e.

        # Given a Chemical system with only Reaction('A <=> B')

        # d[A]dt = -r1 + r2 = -k01[A] + k2[B]

        # being A.key = 0 and B.key = 1

        A = <ElementalStep 1 of Type '=>' with reacts=(1,) products(2,)>

        'dxdt[0]', '-k01*x[1] +k02[B]' = PythonWriter.massbalance(A)

        Parameters
        ----------
        reaction : MassBalance
            Any MassBalance

        Returns
        -------
        var,expr
            variable and expresion of the reaction
        """
        pass
    @abstractmethod
    def jacobian_element(self,Jac_ij):
        """
        Takes the partial differential of a MB and returns the variable of the
        jacobian and the string of the expression. I.e.

        # Given a Chemical system with only Reaction('A <=> B')

        # d[A]dt = -r1 + r2 = -k01[A] + k2[B]

        # being A.key = 0 and B.key = 1

        C = <JacobianElement(Reaction(1),Compound(A))>
        
        'Jac[0,0]', '-k01' = PythonWriter.jacobian_element(A)

        Parameters
        ----------
        Jac_ij : JacobianElement
            Any JacobianElement object

        Returns
        -------
        var,expr
            variable and expresion of the reaction
        """
        pass

    @abstractmethod
    def _function(self,chemicalsys):
        pass
    @abstractmethod
    def _jacobian(self,chemicalsys):
        pass

    # main methods for writing
    def fill_header(self,chemicalsys):
        kwargs = dict()
        kwargs.update(self.parameters)
        for attr in ['dxdt','f','jac', 'jac_f']:
            kwargs[attr] = getattr(self,attr)
        self.header = self._header.format_map(kwargs)
    def fill_tail(self,chemicalsys):
        kwargs = dict()
        kwargs.update(self.parameters)
        for attr in ['dxdt','f','jac', 'jac_f']:
            kwargs[attr] = getattr(self,attr)
        self.tail = self._tail.format_map(kwargs)

    def fill(self,chemicalsys):
        """
        Reads the information of the chemical system and updates the values
        needed for writing.
        """
        self.fill_header(chemicalsys)
        self.fill_tail(chemicalsys)
        self.function = self._function(chemicalsys)
        self.jacobian = self._jacobian(chemicalsys)
    def clear(self):
        """
        Clears the variables related to writing.
        """
        self.header = ''
        self.tail = ''
        self.jacobian = ''
        self.function = ''

    @abstractmethod
    def write(self,chemicalsys,filepath):
        """ Writes the current chemical system into the file specified """
        pass



"""
This module contains several utilities grouped by usage:

   *   The 'Pykinetic Utilities' are functions with usage in the scripts that 
       accompany this library.
   *   Class Specializations. Subclasses of the ChemicalSystem used for the
       scripts pykinetic-model and pykinetic-scan.
"""

import warnings
import math
from itertools import chain

from .classes import ChemicalSystem, Reaction, Energy, DiffusionTS

########################## Pykinetic Utilities #################################
def write_indexfile(chemsys,file,withoutTS=True,isrelative=False):
    """
    Writes a File that summarizes both Compounds and Reactions Files as
    they were modeled. Simple division of this file should lead to two files
    that should be able reproduce at least the function and the Jacobian.

    Parameters
    ----------
    chemsys : ChemicalSystem
        A ChemicalSystem or subclass to output
    FilePath : str
        A valid filepath for the IndexFile
    """
    Out = []
    Out.append('#### compounds ####')
    for compound in chemsys.compounds:
        key, label, energy = compound.key, compound.label, compound.energy
        Out.append(f'{key})    {label}    {energy}')
    Out.append('#### reactions ####')
    if withoutTS:
        for reaction in chemsys.reactions:
            key = reaction.key
            if isrelative:
                energy = reaction.AE.as_unit(chemsys.unit)
            else:
                energy = reaction.TS.energy
            Out.append(f'{key})    {reaction}    !{energy}')
    else:
        for reaction in chemsys.reactions:
            key = reaction.key
            Out.append(f'{key})    {reaction}    !{reaction.TS.label}')
        for TS in chemsys.transitionstates: 
            Out.append(f'{TS.label}    {TS.energy.as_unit(chemsys.unit)}')
    with open(file,'w') as F :
        F.write('\n'.join(Out))
        F.write('\n')
def calc_standard_state_correction(T,pressure='atm'):
    """
    Calculates the standard state correction to 1M and temperature=T. 


    Parameters
    ----------
    T : float
        Temperature in K
    pressure : str, optional
        units of the standard state considered to calculate gibbs free energies,
        Usually it is either 1atm or 1bar, by default 'atm'

    Returns
    -------
    Energy
        Standard state correction to 1M and the specified temperature
    """
    # constants from "The NIST Reference on Constants, Units, and Uncertainty"
    R_SI = 8.314462618    # J/(mol K)
    if pressure == 'bar': 
        R_bar = 0.0831446261815324
        R = R_bar
    else: 
        R_atm = 0.0820573661  # atm L / (mol K)
        R = R_atm
    return Energy(R_SI*T*math.log(R*T),'J/mol')

######################### Class Specializations ################################
class BiasedChemicalSystem(ChemicalSystem):
    """
    A specialization used to directly apply a constant bias to the energy of
    to all the species (as the Standard State Correction does). It is applied to
    compounds and TS except from the diffusion TSs.

    Parameters
    ----------
    bias : float, Energy
        bias to the energy of each species (the default is 0.0).
    bias_unit: str
        (the default is 'kcal/mol').
    T : float
        Temperature in K (the default is 298.15).
    unit : str
        (the default is 'kcal/mol').
    """
    def __init__(self,bias=0.0,bias_unit='kcal/mol',T=298.15,unit='kcal/mol'):
        super().__init__(T,unit)
        if isinstance(bias,Energy):
            self._bias = bias
        else: 
            self._bias = Energy(bias,bias_unit)
    def apply_bias(self):
        for compound in self.compounds: 
            compound.energy = compound.energy + self.bias
        for TS in self.transitionstates: 
            if not hasattr(TS,'barrier'): # only DiffusionTS have .barrier
                TS.energy = TS.energy + self.bias
    def remove_bias(self):
        for compound in self.compounds: 
            compound.energy = compound.energy - self.bias
        for TS in self.transitionstates:
            if not hasattr(TS,'barrier'): # only DiffusionTS have .barrier
                TS.energy = TS.energy - self.bias
    def change_bias(self,bias):
        self.remove_bias()
        if isinstance(bias,Energy):
            self._bias = bias
        else:
            warnings.warn("Guessing energy unit from previous bias' unit")
            self._bias = Energy(bias,self._bias.unit)
        self.apply_bias()
    
    @property
    def bias(self):
        return self._bias
    @bias.setter
    def bias(self,other):
        self.change_bias(other)
class ScannableChemicalSystem(BiasedChemicalSystem):
    """
    A specialization of BiasedChemicalSystem that provides the methods to apply
    dinamically change the energy of all the compounds and TS flagged as 
    scannable.

    Parameters
    ----------
    scan : float, Energy
        bias to the energy of each species (the default is 0.0).
    scan_unit : str
        (the default is 'kcal/mol').
    bias : float, Energy
        bias to the energy of each species (the default is 0.0).
    bias_unit : str
        (the default is 'kcal/mol').
    T : float
        Temperature in K (the default is 298.15).
    unit : str
        (the default is 'kcal/mol').

    """
    def __init__(self,scan=0.0,scan_unit='kcal/mol',
                 bias=0.0,bias_unit='kcal/mol',T=298.15,unit='kcal/mol'):
        super().__init__(bias,bias_unit,T,unit)
        if isinstance(scan,Energy):
            self._scan = scan
            self._scan_unit = scan.unit
        else: 
            self._scan = Energy(scan,scan_unit)
            self._scan_unit = scan_unit

    def apply_scan(self):
        for item in chain(self.compounds,self.transitionstates): 
            if item.scannable:
                if not hasattr(item,'barrier'): 
                    item.energy = item.energy + self.scan
                else: 
                    item.barrier = item.barrier + self.scan
    def remove_scan(self):
        for item in chain(self.compounds,self.transitionstates): 
            if item.scannable:
                if not hasattr(item,'barrier'): 
                    item.energy = item.energy - self.scan
                else: 
                    item.barrier = item.barrier - self.scan
    def change_scan(self,scan):
        self.remove_scan()
        if isinstance(scan,Energy):
            self._scan = scan
        else:
            self._scan = Energy(scan,self._scan_unit)
        self.apply_scan()
    
    @property
    def scan(self):
        return self._scan
    @scan.setter
    def scan(self,other):
        self.change_scan(other)




from pykinetic.classes import *
import unittest

class CompoundTest(unittest.TestCase):
    def setUp(self):
        self.label1 = 'label1'
        self.label2 = 'label2'
        self.energy1 = 0.0
        self.energy2 = 5.0
        self.energy3 = -5.0
        self.key1 = 0
        self.key2 = 'Key2'

    def test_init(self):
        C0 = Compound(self.label1,self.energy1,self.key1,True)
        C1 = Compound(self.label2,self.energy2)
        self.assertEqual(C0.label, self.label1,'Error constructing object')
        self.assertEqual(C0.energy,self.energy1,'Error constructing object')
        self.assertEqual(C0.key,self.key1,'Error constructing object')
        self.assertEqual(C1.key,None,'Error constructing object')
        self.assertTrue(C0.scannable,'Error constructing object')
        self.assertFalse(C1.scannable,'Error constructing object')

    # Test Logic
    def test_equal(self):
        C0 = Compound(self.label1,self.energy1,self.key1)
        C1 = Compound(self.label2,self.energy2)
        C2 = Compound(self.label1,self.energy3,self.key2)
        self.assertEqual(C0,C2)
        self.assertEqual(C2,C0)
        self.assertEqual(C0 == C1,False)
        self.assertEqual(C2 == C1,False)
        self.assertEqual(C1 == self.label2,False)

    # Test Hashing
    def test_hash(self):
        C0 = Compound(self.label1,self.energy1,self.key1)
        C1 = Compound(self.label2,self.energy2)
        C2 = Compound(self.label1,self.energy3,self.key2)
        C3 = Compound(self.label1,self.energy3,self.key1)
        self.assertEqual(hash(C0),hash(C0))
        self.assertNotEqual(hash(C0),hash(C1))
        self.assertEqual(hash(C0),hash(C2))
        self.assertEqual(hash(C0),hash(C3))

    # Test number conversion
    def test_int(self):
        C0 = Compound(self.label1,self.energy1,self.key1)
        C1 = Compound(self.label2,self.energy2)
        C2 = Compound(self.label1,self.energy3,self.key2)
        C3 = Compound(self.label1,self.energy3,self.key1)
        self.assertEqual(int(C0),int(self.key1))
        self.assertEqual(int(C0),int(C3))
        with self.assertRaises(ValueError):
            _ = int(C2)
        with self.assertRaises(TypeError):
            _ = int(C1)

class ReactionTest(unittest.TestCase):
    def setUp(self):
        self.Compound1 = Compound('C1',Energy(0.0,'kcal/mol'))
        self.Compound2 = Compound('C2',Energy(0.0,'kcal/mol'))
        self.Compound3 = Compound('C3',Energy(-2.0,'kcal/mol'))
        self.TS = TransitionState(Energy(10.0,'kcal/mol'),label='TS1')

    def test_init(self):
        reactants = (self.Compound1, self.Compound2)
        products  = (self.Compound3,)
        reaction = Reaction(reactants,products)
        self.assertTrue(reaction.reactants[self.Compound1],1)
        self.assertTrue(reaction.reactants[self.Compound2],1)
        self.assertTrue(reaction.products[self.Compound3],1)
        self.assertIsNone(reaction.key)
        self.assertIsNone(reaction.TS)

    def test_coefficient(self):
        reactants = (self.Compound1, self.Compound1, self.Compound2)
        products  = (self.Compound3, self.Compound1)
        reaction = Reaction(reactants,products)
        self.assertEqual(reaction.coefficient(self.Compound1),-1)
        self.assertEqual(reaction.coefficient(self.Compound2),-1)
        self.assertEqual(reaction.coefficient(self.Compound3),1)

    # Test Hashing
    def test_properties(self):
        reactants = (self.Compound1, self.Compound2)
        products  = (self.Compound3,)
        reaction = Reaction(reactants,products)
        reaction.TS = self.TS
        with self.subTest(property='reactants_energy'):
            self.assertEqual(reaction.reactants_energy,Energy(0.0,'kcal/mol'))
        with self.subTest(property='AE'):
            self.assertEqual(reaction.AE,Energy(10.0,'kcal/mol'))
            reaction.AE = Energy(0.0,'J/mol')
            self.assertEqual(reaction.AE,Energy(0.0,'J/mol'))
        with self.subTest(property='k'):
            k = 298.15*1.38064852E-023/6.62607004E-034
            self.assertAlmostEqual(k,reaction.k)

    # Test number conversion
    def test_errors(self):
        reactants = (self.Compound1, self.Compound2)
        products  = (self.Compound3,)
        reaction = Reaction(reactants,products)
        with self.assertRaises(RuntimeError):
            reaction.AE = 10
        reaction.TS = self.TS
        with self.assertRaises(ValueError):
            reaction.AE = -1.0
            _ = reaction.k

class MassBalanceTest(unittest.TestCase):
    def setUp(self):
        C1 = self.Compound1 = Compound('C1',Energy(0.0,'kcal/mol'))
        C2 = self.Compound2 = Compound('C2',Energy(0.0,'kcal/mol'))
        C3 = self.Compound3 = Compound('C3',Energy(-2.0,'kcal/mol'))
        self.reaction1 = Reaction(reactants=(C1,C2),products=(C3,))
        self.reaction2 = Reaction(reactants=(C3,),products=(C1,C2))
        self.reaction3 = Reaction(reactants=(C1,C1),products=(C1,C3))

    def test_partial(self):
        C1 = Compound('C1',Energy(0.0,'kcal/mol'))
        C2 = Compound('C2',Energy(0.0,'kcal/mol'))
        C3 = Compound('C3',Energy(-2.0,'kcal/mol'))
        r1 = Reaction(reactants=(C1,C2),products=(C3,))
        r2 = Reaction(reactants=(C3,),products=(C1,C2))
        r3 = Reaction(reactants=(C1,C1),products=(C1,C3))
        r1.key = 1
        r2.key = 2 
        r3.key = 3
        MB1 = MassBalance(C1,items=[(-1,r1),(1,r2),(-1,r3)])
        MB2 = MassBalance(C2,items=[(-1,r1),(1,r2)])
        MB3 = MassBalance(C3,items=[(1,r1),(-1,r2),(1,r3)])
        partials = [[JacobianElement(C1,C1,items=[(-1,r1),(1,r2),(-1,r3)]),
                     JacobianElement(C1,C2,items=[(-1,r1),(1,r2)]),
                     JacobianElement(C1,C3,items=[(-1,r1),(1,r2),(-1,r3)])],
                    [JacobianElement(C2,C1,items=[(-1,r1),(1,r2)]),
                     JacobianElement(C2,C2,items=[(-1,r1),(1,r2)]),
                     JacobianElement(C2,C3,items=[(-1,r1),(1,r2)])],
                    [JacobianElement(C3,C1,items=[(1,r1),(-1,r2),(1,r3)]),
                     JacobianElement(C3,C2,items=[(1,r1),(-1,r2)]),
                     JacobianElement(C3,C3,items=[(1,r1),(-1,r2),(1,r3)])]]
        for i,(MB,elements) in enumerate(zip([MB1,MB2,MB3],partials)): 
            for j,(C,element) in enumerate(zip([C1,C2,C3],elements)):
                with self.subTest(MB=i+1,partial=j+1):
                    test = MB.partial(C).items
                    self.assertEqual(test,element.items) 

class EnergyTest(unittest.TestCase):
    def setUp(self):
        u1,u2 = 'kcal/mol','kJ/mol'
        factors = Energy._ConversionFactors
        self.zero = Energy(0.0,u1)  # zero
        self.pos1 = Energy(1.0,u1) # positive
        self.pos2 = Energy(2.0,u1) # higher than pos1
        self.neg1 = Energy(-1.0,u1) # negative
        self.neg2 = Energy(-2.0,u1) # lower than neg1
        self.refpos = Energy(1.0/factors[u1])
        self.refneg = Energy(-1.0/factors[u1])
        val = factors[u2]/factors[u1]
        self.unitpos = Energy(val,u2)
        self.unitneg = Energy(-val,u2)

    def test_init(self):
        e0 = Energy(unit='kcal/mol')
        e1 = Energy( 1.2,'kcal/mol')
        e2 = Energy(-6.3)
        e3 = Energy(e0,'kJ/mol')
        self.assertEqual(e0.value, 0.0,'Error initializing default value')
        self.assertEqual(e1.value, 1.2,'Error constructing object')
        self.assertEqual(e1._unit,'kcal/mol','Error constructing object')
        self.assertEqual(e2._unit,'hartree','Error initializing default unit')
        self.assertEqual(e3.value == e0.value, True)
        self.assertEqual(e3.unit == e0.unit, False)
    # Test Properties
    def test_value(self):
        compare = lambda x,y: abs(x-y)<=1E-12
        Test = Energy(5.0,'kcal/mol')
        self.assertEqual(compare(Test.value,5.0),True)
        self.assertEqual(compare(Test._value,5.0/Test._conversion),True)
        Test.value = 0.0
        self.assertEqual(compare(Test._value,0.0),True)
        Test._value = 5.0/Test._conversion
        self.assertEqual(compare(Test.value,5.0),True)

    def test_unit(self):
        Test = Energy(5.0,'kcal/mol')
        self.assertEqual(Test.unit,'kcal/mol')
        self.assertEqual(Test._unit,'kcal/mol')
        Val = Test._value
        Test.unit = 'hartree'
        self.assertEqual(Test.value,Val)
        self.assertEqual(Test.unit,'hartree')
        self.assertEqual(Test._unit,'hartree')
        with self.assertRaises(NotImplementedError):
            Test.unit = 'FakeUnit'

    # Test Logic
    def test_bool(self):
        self.assertEqual(bool(self.zero), False)
        self.assertEqual(bool(self.pos1), True)
    def test_equal(self):
        self.assertEqual(self.pos1 == self.unitpos, True)
        self.assertEqual(self.zero == self.pos1, False)
        self.assertEqual(self.pos1 == self.pos1.value, True)
        self.assertEqual(self.pos1 == self.zero.value,False)
    def test_gt(self):
        self.assertEqual(self.pos2 > self.pos1, True)
        self.assertEqual(self.pos2 > self.pos1.value, True)
        self.assertEqual(self.pos2.value > self.pos1, True)
        self.assertEqual(self.neg1 > self.zero.value, False)
        self.assertEqual(self.neg1 > self.neg2,True)
    def test_lt(self):
        self.assertEqual(self.pos2 < self.pos1, False)
        self.assertEqual(self.pos2 < self.pos1.value, False)
        self.assertEqual(self.pos2.value < self.pos1, False)
        self.assertEqual(self.neg1 < self.zero.value, True)
        self.assertEqual(self.neg1 < self.neg2,False)
    def test_ne(self):
        self.assertEqual(self.pos1 != self.pos2, True)
        self.assertEqual(self.pos1 != abs(self.neg1),False)
        self.assertEqual(self.pos1.value != self.pos2, True)

    # Test number conversion
    def test_int(self):
        self.assertEqual(int(self.pos1),int(self.pos1.value))
    def test_float(self):
        self.assertEqual(float(self.pos1),self.pos1.value)
    def test_round(self):
        self.assertEqual(round(self.pos1,5),round(self.pos1.value,5))

    # Test Math
    def test_add_1(self):
        #Test if math addition works properly
        Out = [ (self.pos1+self.zero,self.pos1), # + 0
                (self.neg1+self.zero,self.neg1), # - 0
                (self.pos1+self.pos1,self.pos2), # + +
                (self.pos1+self.neg1,self.zero), # + -
                (self.neg1+self.neg1,self.neg2)] # - -
        for test,result in Out:
            self.assertEqual(test.value,result)
    def test_add_2(self):
        # Test effect of different units and ensure Energy Class Output
        Items = [self.pos1+self.unitpos,
                self.pos1+self.refpos,
                self.unitpos+self.pos1,
                self.pos1+self.pos1.value]
        Out = [ (Items[0],self.pos2, True,self.pos1.unit),
                (Items[1],self.pos2, True,self.pos1.unit),
                (Items[2],self.pos2, True,self.unitpos.unit),
                (Items[3],self.pos2, True,self.pos1.unit)]
        for item,test,result,unit in Out:
            self.assertEqual(item,test)
            self.assertEqual(item.unit,unit)
            self.assertIsInstance(item,Energy)

    def test_sub_1(self):
        #Test if math substraction works properly
        Out = [ (self.zero-self.pos1,self.neg1), # 0 +
                (self.zero-self.neg1,self.pos1), # 0 -
                (self.pos1-self.pos1,self.zero), # + +
                (self.pos1-self.zero,self.pos1), # + -
                (self.pos1-self.neg1,self.pos2)] # - -
        for test,result in Out:
            self.assertEqual(test.value,result)
    def test_sub_2(self):
        # Test effect of different units and ensure Energy Class Output
        Out = [ (self.pos1-self.unitpos,self.zero,self.pos1.unit),
                (self.pos1-self.refpos,self.zero,self.pos1.unit),
                (self.unitpos-self.pos1,self.zero,self.unitpos.unit),
                (self.pos1-self.pos1.value,self.zero,self.pos1.unit)]
        for test,result,unit in Out:
            self.assertEqual(test.value,result)
            self.assertEqual(test.unit,unit)
            self.assertIsInstance(test,Energy)

    def test_mul(self):
        self.assertEqual(self.pos1*2.0,self.pos2)
        self.assertEqual(self.pos1*(-1.0),self.neg1)
        self.assertEqual(self.zero*2.0,self.zero)
        with self.assertRaises(NotImplementedError):
            self.zero*self.zero

    def test_div_1(self):
        # Test correct behaviour of division
        self.assertEqual(self.pos2/2.0,self.pos1)
        self.assertEqual(self.pos1/(-1.0),self.neg1)
        self.assertEqual(self.zero/2.0,self.zero)
        self.assertEqual(self.pos1/self.pos1,1.0)
    def test_div_2(self):
        # Test instance/unit consistency
        test1 = self.pos1/self.unitpos
        test2 = self.pos1/2.0
        self.assertIsInstance(test1,float)
        self.assertIsInstance(test2,Energy)
        self.assertEqual(test2.unit,self.pos1.unit)
        self.assertEqual(test2.value,self.pos1.value/2.0)

    def test_neg(self):
        self.assertEqual(-self.pos1,self.neg1)
        self.assertEqual(-self.neg1,self.pos1)
        self.assertEqual(-self.zero,self.zero)

    def test_abs(self):
        self.assertEqual(abs(self.pos1),self.pos1)
        self.assertEqual(abs(self.neg1),self.pos1)
        self.assertEqual(abs(self.zero),self.zero)

    def test_inline(self):
        # Test +=, -=, *= and /=
        Test = Energy(self.zero.value,'kcal/mol')
        val = self.pos1.value
        Test += val
        self.assertEqual(Test.value,val,'incorrect iadd implementation')
        Test -= val
        self.assertEqual(Test == self.zero, True,'incorrect isub implementation')
        Test += self.pos2
        self.assertEqual(Test,self.pos2,'incorrect iadd implementation')
        Test -= self.pos2
        self.assertEqual(Test == self.zero, True,'incorrect isub implementation')
        Test += self.pos1
        Test *= 0.5
        self.assertEqual(Test,self.pos1*0.5,'incorrect imul implementation')
        Test /= 0.5
        self.assertEqual(Test,self.pos1,'incorrect itruediv implementation')
        Test /= self.pos1
        self.assertEqual(Test,1.0,'incorrect itruediv implementation')

    def test_commutative_eq(self):
        self.assertEqual(self.zero == self.pos1, self.pos1 == self.zero)
    def test_commutative_add(self):
        self.assertEqual(self.pos1 + self.zero, self.zero + self.pos1)
        self.assertEqual(self.pos1 + 1.0, 1.0 + self.pos1)
    def test_commutative_sub(self):
        self.assertEqual(self.pos1 - self.pos2, -self.pos2 + self.pos1)
        self.assertEqual(self.pos1 - 1.0, -1.0 - (-self.pos1))
    def test_commutative_mul(self):
        self.assertEqual(self.pos1*2.0,2.0*self.pos1)

    # Test Methods
    def test_as_unit(self):
        val,unit = self.pos1.value, self.pos1.unit
        unit2 = self.unitpos.unit
        Out1 = self.pos1.as_unit(unit2)
        self.assertIn(str(self.unitpos.value),Out1)
        self.assertIn(unit2,Out1)
        with self.assertRaises(NotImplementedError):
            self.pos1.as_unit('FakeUnit')



"""
This module contains the basic functions to read the user provided files and 
populate a ChemicalSystem class from them. The core function of this module is
'populate_chemicalsystem_fromfiles'.
"""

import re 
from itertools import chain

from .classes import Compound,Energy,Reaction,TransitionState,DiffusionTS


# Custom errors
class MissingCompounds(ValueError):
    pass
class MissingTransitionStates(ValueError):
    pass

# GLOBALS
DIFFUSION = '<d>'
BACKWARDS = '<='
FORWARD = '=>'
EQUILIBRIUM = '<=>'
MARKERS = {EQUILIBRIUM: {'TS':TransitionState,},  # Reversible
           FORWARD : {'TS':TransitionState,},  # Direct reaction
           BACKWARDS : {'TS':TransitionState,},  # Backwards reaction
           DIFFUSION: {'TS':DiffusionTS,}}      # Reversible diffusion

REVERSIBLE = [EQUILIBRIUM,DIFFUSION] # reactions modeled as 2 elemental steps

for mark in MARKERS: 
    MARKERS[mark]['matcher'] = re.compile(f'(.*)\s{mark}\s(.*)\s!(.*)')

re_is_energy = re.compile('(-?[0-9]*\.+[0-9]*)\s*?([^\s]*?)')
def is_energy(text):
    """
    Returns if a given string corresponds to an energy or not. If the string is 
    a 

    Parameters
    ----------
    text : str
        A string that may be an energy

    Returns
    -------
    bool or truth-like value
        If True or True-like then the evaluated string is an energy.
    """
    test = re_is_energy.findall(text)
    if test:
        return test[0]
    else:
        return False

def populate_chemicalsystem_fromfiles(chemicalsystem,file_c,file_r,
                                      energy_unit='J/mol',relativeE=False):
    """
    Fills a chemicalsystem with the compounds and reactions specified in the 
    compunds file and the reactions file. 

    Parameters
    ----------
    chemicalsystem : pykinetic.classes.ChemicalSystem
        chemical system that will be populated
    file_c : str or Path
        path to the correctly formatted compounds file  
    file_r : str or Path
        path to the correctly formatted reactions file
    energy_unit : str, optional
        default unit to assume when none is found, by default 'J/mol'
    relativeE : bool, optional
        If true the energies of the reactions are assumed to be the barriers 
        relative to reactants instead of assumed to be the energy of the 
        transition state connecting reactants and products, by default False
    """

    # Process the Compound File
    raw_compounds = read_compounds(file_c)
    compounds = create_compounds(raw_compounds,energy_unit)
    chemicalsystem.cextend(compounds)

    # Process the reaction File
    raw_reactions, TS_lines = read_reactions(file_r)
    if TS_lines: 
        TS_dict = create_TS_dict(TS_lines,energy_unit)
    else:
        TS_dict = dict()

    check_missing_compounds(chemicalsystem,raw_reactions)
    check_missing_TS(raw_reactions,TS_dict)

    # Parse the reactions and create the reaction and ts instances
    for reactants,mark,products,TS_text in raw_reactions: 
        reactants = [chemicalsystem.Name2Compound[r] for r in reactants]
        products  = [chemicalsystem.Name2Compound[p] for p in products]
        reactants_energy = sum([r.energy for r in reactants])
        if mark == BACKWARDS:
            reactants, products = products, reactants
        label,energy,scannable = prepare_inline_TS(TS_text,TS_dict,energy_unit)
        # Create the reaction/s
        reactions = [Reaction(reactants,products),]
        if mark in REVERSIBLE:
            reactions.append(Reaction(products,reactants))
        # Handle relative energy specification
        if mark != DIFFUSION and relativeE and label not in TS_dict: 
            energy = energy + reactants_energy
        # Create the TS
        ts_cls = MARKERS[mark]['TS']
        if label not in chemicalsystem.Name2TS: 
            TS = ts_cls(energy,reactions=reactions,label=label,
                        scannable=scannable)
        else:
            TS = chemicalsystem.Name2TS[label]
            TS.reactions.extend(reactions)
        # Add the reactions to the chemical system
        for reaction in reactions: 
            reaction.TS = TS
            chemicalsystem.radd(reaction,False)
        chemicalsystem.rupdate()

def read_compounds(file):
    """
    To-document
    """
    raw_compounds = []
    with open(file,"r") as F:
        for line in F:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            raw_compounds.append(line.split())
    return raw_compounds
def create_compounds(raw_compounds,energy_unit='J/mol'):
    """
    To-document
    """
    err_msg = "Unexpected number of items in \n '{}' "
    compounds = []
    for items in raw_compounds:
        scannable = 'scan' in items[-1]
        if scannable:
            _ = items.pop(-1)
        if len(items) == 3:
            label,energy,unit = items
        elif len(items) == 2: 
            label,energy = items
            unit = energy_unit
        else:
            raise RuntimeError(err_msg.format('  '.join(items)))
        compound = Compound(label,Energy(energy,unit),scannable=scannable)
        compounds.append(compound)
    
    try: # Check if there is any duplicates
        assert len(set(compounds)) == len(compounds)
    except AssertionError as e: 
        msg = 'Inconsistent number of compounds. Check for duplicates'
        raise ValueError(msg)
    return compounds

def read_reactions(file):
    """
    To-document
    """
    reaction_lines = []
    TS_lines = [] 
    with open(file,'r') as F:
        lines = F.readlines()
    for line in lines:
        line = line.strip()
        if not line or line.startswith('#'):
            pass
        elif '!' in line:   # [A-]  +  [B+] => C  !TS1 
            reaction_lines.append(line)
        else:               # TSNAME    Number   (Unit)
            TS_lines.append(line)
    raw_reactions = [split_reaction_line(line) for line in reaction_lines]
    return raw_reactions,TS_lines
def create_TS_dict(TS_lines,energy_unit='J/mol'):
    """
    To-document
    """
    TS_dict = dict()
    for line in TS_lines:
        scannable = False
        items = line.strip().split()
        # Guess if it is a scannable TS
        if items[-1] == 'scan':
            scannable = True
            _ = items.pop(-1)
        # now only two cases are possible
        if len(items) == 3:  # Energy unit is specified
            label,energy,unit = items
        elif len(items) == 2: # energy unit is not specified
            label, energy = items
            unit = energy_unit
        else:
            raise RuntimeError(f"Unexpected number of items in \n '{line}' ")
        TS_dict[label] = (Energy(energy,unit),scannable)
    
    try: # Check if there is any duplicates
        assert len(TS_dict) == len(TS_lines)
    except AssertionError as e: 
        raise ValueError("Inconsistent number of TSs at the end of the file. " \
                         "Check for duplicates")
    return TS_dict
def prepare_inline_TS(TS_text,TS_dict,energy_unit='J/mol'):
    """
    To-document
    """
    scannable = False
    # Prepare the TS data
    TS_items = TS_text.split()
    if len(TS_items) == 1 and not is_energy(TS_text):
        label = TS_items[0]
        energy,scannable = TS_dict[label]
    elif len(TS_items) == 1: 
        label = None
        energy = Energy(TS_items[0],energy_unit)
    elif len(TS_items) == 2:
        label = None
        number, unit = TS_items
        energy = Energy(number,unit)
    elif len(TS_items) == 3 and TS_items[-1] == 'scan':
        label = None
        number, unit = TS_items[:2]
        energy = Energy(number,unit)
        scannable = True
    else: 
        ts = ' '.join(TS_items)
        raise ValueError(f'Wrong TS definition at "{ts}"')
    return label,energy,scannable
def split_reaction_line(reaction_text): 
    """
    To-document
    """
    for mark in MARKERS: 
        match = MARKERS[mark]['matcher'].findall(reaction_text)
        if match:
            break
    else:
        raise ValueError('the reaction does not contain a known marker')

    reactants_text, products_text, TS_text = match[0]
    reactants = [r.strip() for r in reactants_text.split(' + ')]
    products  = [p.strip() for p in products_text.split(' + ')]
    
    return  reactants, mark, products, TS_text.strip()

def check_missing_compounds(chemicalsystem,raw_reactions):
    """
    Checks if all the compounds in the chemicalsystem are 'defined' 
    within the ChemicalSystem instance.

    Raises
    ------
    MissingCompounds

    """
    defined_compounds = set(chemicalsystem.Name2Compound)
    reactants,mark,products,TS_text = zip(*raw_reactions)
    reaction_compounds = set(list(chain(*reactants)) + list(chain(*products)))
    missing_compounds = reaction_compounds - defined_compounds
    if missing_compounds:
        msg = 'The following compounds are not defined: \n{}\n'
        items = [f'{item}\tNUMBER\tUNIT?' for item in sorted(missing_compounds)]
        raise MissingCompounds(msg.format('\n'.join(items)))
def check_missing_TS(raw_reactions,TS_dict):
    """
    Checks if all the labeled transition states of the reactions are 'defined' 
    are defined.

    Raises
    ------
    MissingTransitionStates

    """
    defined_TS = set(TS_dict)
    reactants,mark,products,TS_text = zip(*raw_reactions)
    reaction_TS = []
    for item in TS_text :
        if not is_energy(item): 
            reaction_TS.append(item.split()[0])
    missing_TS = set(reaction_TS) - defined_TS
    if missing_TS:
        msg = 'The following labeled TS are not defined: \n{}\n'
        items = [f'{item}\tNUMBER\tUNIT?\tSCAN?' 
                 for item in sorted(missing_TS)]
        raise MissingTransitionStates(msg.format('\n'.join(items)))


"""
This module contains the core classes of the library: Energy, 
MassBalance, JacobianElement, Compound, TransitionState, DiffusionTS, Reaction, 
Chemical System and SimulationParameters and ConvergenceParameters. 
"""

from itertools import chain
from collections import Counter, UserDict
from abc import abstractmethod
import warnings

import numpy

class Energy(object):
    """
    Objects that represents a number with an energy unit.

    Parameters
    ----------
    value : float
        Number or object with a __float__ (the default is 0.0).
    unit : str
        Energy unit in which the value parameter is provided. See class variable
        _units for the different implemented units (the default is 'hartree').

    Attributes
    ----------
    _units : list
        {}
    unit
    value

    """
    _units = 'hartree eV cm-1 kcal/mol kJ/mol J/mol K J Hz'.split()
    __doc__ = __doc__.format(_units)
    _Factors = (1.0, 27.2107, 219474.63, 627.503, 2625.5, 2625500.0,
                315777.0, 43.60E-19, 6.57966E15)
    _ConversionFactors = {key:val for key,val in zip(_units,_Factors)}

    __slots__ = ('_value','_conversion','_SI','_unit')

    def __init__(self,value=0.0,unit='hartree'):
        # Assume that value is in the specified unit
        self._conversion = self._ConversionFactors[unit]
        self._SI = self._ConversionFactors['J/mol']
        self.unit = unit
        self.value = float(value)
    def __repr__(self):
        cls = self.__class__.__name__
        return f'<{cls}:{self.value},{self.unit}>'
    def __str__(self):
        return f'{self.value} {self.unit}'
    def __format__(self,format_spec):
        return ' '.join([self.value.__format__(format_spec),self.unit])
    def __bool__(self):
        return self.value.__bool__()

    @property
    def value(self):
        return self._value * self._conversion
    @value.setter
    def value(self,other):
        self._value = float(other)/self._conversion

    @property
    def unit(self):
        return self._unit
    @unit.setter
    def unit(self,other):
        try:
            self._conversion = type(self)._ConversionFactors[other]
        except KeyError:
            raise NotImplementedError('Unit {} not implemented'.format(other))
        self._unit = other

    # Emulate Numeric Type
    def __add__(self,other):
        """ Implements the 'self + other' operation """
        Out = Energy(unit=self.unit)
        try:
            Out._value =  self._value + other._value
        except AttributeError:
            Out.value = self.value + other
        finally:
            return Out
    def __sub__(self,other):
        """ Implements the 'self - other' operation """
        return self.__add__(-other)
    def __mul__(self,other):
        """ Implements the 'self * other' operation """
        if isinstance(other,type(self)):
            raise NotImplementedError('Energy*Energy not implemented')
        Out = Energy(self.value*other,unit=self.unit)
        return Out
    def __div__(self,other):
        """ Defined for compatibility with Python2.7, Implements the
        'self/other' behaviour in python 2 """
        return self.__truediv__(other)
    def __truediv__(self,other):
        """ Implements the 'self / other' operation """
        Out = Energy(unit=self.unit)
        try:
            _ = other._value
        except AttributeError:
            Out.value = self.value/other
        else:
            Out = self._value/other._value
        return Out
    def __radd__(self,other):
        """ Implements the 'other + self' operation """
        return self + other
    def __rsub__(self,other):
        """ Implements the 'self - other' operation """
        return self.__add__(-other)
    def __rmul__(self,other):
        return self * other
    def __iadd__(self,other):
        """ Implements the '+= other' operation """
        try:
            self._value = self._value + other._value
        except AttributeError:
            self.value = self.value + other
        return self
    def __isub__(self,other):
        """ Implements the '-= other' operation """
        self.__iadd__(-other)
        return self
    def __imul__(self,other):
        """ Implements the '*= other' operation """
        self._value = self._value * other
        return self
    def __itruediv__(self,other):
        """ Implements the '/= other' operation """
        Out = self
        try:
            _ = other._value
        except AttributeError:
            self._value = self._value / other
        else:
            Out = self._value / other._value
        finally:
            return Out

    def __neg__(self):
        """ Implements the '- self' operation """
        return Energy(-self.value,self.unit)
    def __abs__(self):
        """ Implements the abs() function behaviour """
        return Energy(abs(self.value),self.unit)
    def __int__(self):
        """ Implements the int() function behaviour """
        return int(self.value)
    def __float__(self):
        """ Implements the float() function behaviour """
        return float(self.value)
    def __round__(self,ndigits):
        """ Implements the round() function behaviour,
        Must return an integrer """
        return round(self.value,ndigits)
    # Comparison methods
    def __eq__(self,other):
        """ Implements the 'self == other' behaviour"""
        # Implementation of equality test extracted from math.isclose
        atol = 1E-6
        rtol = 1E-9
        try:
            A,B = self._value, other._value
        except AttributeError:
            A,B = self.value, other
        test = max(rtol*max(abs(A),abs(B)),atol)
        if abs(A - B) == test:
            warnings.warn('Difference between numbers is close to tolerance')
        return abs(A - B) <= test
    def __ne__(self,other):
        """ Implements the 'self != other' behaviour"""
        return not(self == other)
    def __lt__(self,other):
        """ Implements the 'self < other' behaviour"""
        try:
            Out = self._value < other._value
        except AttributeError:
            Out = self.value < other
        return Out
    def __gt__(self,other):
        """ Implements the 'self > other' behaviour"""
        return not ((self < other) or (self == other))
    def __le__(self,other):
        """ Implements the 'self <= other' behaviour"""
        return not self > other
    def __ge__(self,other):
        """ Implements the 'self >= other' behaviour"""
        return not self < other


    def as_unit(self,unit):
        """Returns a string with the value expressed in the unit specified.

        Parameters
        ----------
        unit : str
            A valid string representing an energy unit, see `Energy._units`.

        Returns
        -------
        str

        """
        try:
            F = type(self)._ConversionFactors[unit]
        except KeyError:
            raise NotImplementedError('Unit {} not implemented'.format(unit))
        Val = str(self._value * F)
        return '{} {}'.format(Val,unit)
    def to_SI(self):
        """Returns the value in units of the SI.

        Returns
        -------
        float

        """
        return self._value*self._SI
    def to_unit(self,unit):
        """Changes the unit of the object.

        Parameters
        ----------
        unit : str
            A valid string representing an energy unit, see `Energy._units`.

        """
        self.unit = unit

class MassBalance(object):
    """
    Representation of a mass balance.

    Parameters
    ----------
    compound : Compound
        Compound of the targeted mass balance
    items : list
        list of Reactions with a net generation/consumption of 'compound' != 0.
    """
    def __init__(self,compound,items=None):
        self.compound = compound
        if items is None:
            self.items = []
        else:
            self.items = items
    def partial(self,compound):
        """
        Do the partial derivative of the mass balance with respect to 'compound'
        and return a JacobianElement object.

        Parameters
        ----------
        compound : Compound
            Compound with which the partial derivative is to be carried out.

        Returns
        -------
        JE
            JacobianElement
        """
        JE = JacobianElement(self.compound,compound)
        for coef,reaction in self.items:
            if compound in reaction:
                JE.items.append((coef,reaction))
        return JE
class JacobianElement(object):
    """
    Representation of an element of a jacobian matrix of the mass balances.

    Parameters
    ----------
    compound1 : Compound
        Compound of the targeted mass balance
    compound2 : Compound
        Compound of the partial derivative.
    items : list
        list of Reactions in which compound2 appears.

    """
    def __init__(self,compound1,compound2,items=None):
        self.compound1 = compound1
        self.compound2 = compound2
        if items is None:
            self.items = []
        else:
            self.items = items

class Compound(object):
    """
    Object representation of a chemical species.

    Parameters
    ----------
    label : str
        identifier/name of the species
    energy : energy
        relative or absolute free energy of the species.
    key : int
        integrer used to represent the species in the overall chemical system
        see `ChemicalSystem` (the default is None).
    scannable : bool
        Flag to identify if this compound's energy is to be modified.

    """

    __slots__ = ('label','energy','key','scannable')

    def __init__(self,label,energy,key=None,scannable=False):
        self.label = label
        self.energy = energy
        self.key = key
        self.scannable = scannable
    def __eq__(self,other):
        try:
            out = self.label == other.label
        except AttributeError:
            out = False
        finally:
            return out
    def __hash__(self):
        return hash(self.label)
    def __int__(self):
        return int(self.key)
    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}('{self.label}','{self.energy}')"
    def copy(self):
        """
        Creates a copy of the object with the key set as None.
        """
        return self.__class___(self.label,self.energy,None,self.scannable)

class TransitionState(object):
    """
    Object representation of the TS. Without an instance of this class assigned
    to a reaction it is not possible to calculate the Activation energy.

    Parameters
    ----------
    energy : Energy
        free energy of the transition state.
    reactions : list
        A list of reactions where the TS is used. It is used to have a reference
        to those instances in case it is needed.
    label : str
        A string to identify the Transition State. If none is provided a default
        one is assigned upon initialization.
    scannable : bool
        Marks if the Transition State is object of a scan.
    """
    defaultname = 0 
    def __init__(self,energy,reactions=None,label=None,scannable=False):
        if label is None:
            label =  f'TS{self.defaultname:02d}'
            self.__class__.defaultname += 1
        self.label = label
        self.energy = energy
        if reactions is None:
            reactions = []
        self.reactions = reactions
        self.scannable = scannable
    def __eq__(self,other):
        try:
            out = self.label == other.label
        except AttributeError:
            out = False
        finally:
            return out
    def __hash__(self):
        return hash(self.label)
    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls} <'{self.label}'>"
    
    @property
    def energy(self):
        return self._energy
    @energy.setter
    def energy(self,other):
        self._energy = other

class DiffusionTS(TransitionState):
    """
    Object representation of the TS for a diffusion process. It requires a 
    reactions argument and a relative barrier. The transition state will always 
    be 'barrier' in energy higher than the highest reactants.

    Parameters
    ----------
    energy : Energy
        free energy of the transition state.
    reactions : list
        A list of reactions where the TS is used. It is used to ensure that 
        any modification to the TS energy or to the compounds of those reactions 
        does not lead t below any of the reactants in those reactions
    label : str
        A string to identify the Transition State. If none is provided a default
        one is assigned upon initialization.
    scannable : bool
        Marks if the Transition State is object of a scan.
    """
    defaultname = 0
    def __init__(self,barrier,reactions,label=None,scannable=False):
        self.barrier = barrier
        if label is None:
            label = f'DiffTS{self.defaultname:02d}'
            self.__class__.defaultname += 1
        self.label = label
        self.reactions = reactions
        self.scannable = scannable
    
    @property
    def energy(self):
        energy = max([r.reactants_energy for r in self.reactions])
        return self.barrier + energy
    @energy.setter
    def energy(self,other):
        energy = max([r.reactants_energy for r in self.reactions])
        if other < energy: 
            raise ValueError("Cannot set an energy lower than the reactants'")
        self.barrier = other - energy

class Reaction(object):
    """
    Object that represents a reaction as an elemental step.

    Parameters
    ----------
    reactants : list
        `Compound` instances that represent the reactants of the reaction
    products : list
        `Compound` instances that represent the products of the reaction
    TS : TransitionState
        Transition state that connects reactants and products. It must contain
        a 'energy' attribute

    Attributes
    ----------
    reactants
    compounds
    key : int
        number that identifies the reaction in a `ChemicalSystem` instance.
    products
    TS
    T : float
        Temperature in K. 
    AE : Energy
        Activation energy defined as AE = energy(TS)-energy(reactants) 
    k : float 
        kinetic constant calculated at the Temperature T.

    """

    def __init__(self,reactants,products,TS=None,T=298.15):
        self.reactants = Counter(reactants)
        self.products = Counter(products)
        self.compounds = {i:0 for i in chain(reactants,products)}
        for i in self.reactants.elements():
            self.compounds[i] -= 1
        for i in self.products.elements():
            self.compounds[i] += 1
        self.key = None
        self.T = T # Temperature in K
        self.TS = TS # Transition state associated, may be shared
    def __str__(self):
        reactants = '  +  '.join([r.label for r in self.reactants.elements()])
        products = '  +  '.join([p.label for p in self.products.elements()])
        out = '{}    {}    {}'.format(reactants,'=>',products)
        return out
    def __repr__(self):
        n = len(self.reactants)
        m = len(self.products)
        i = self.key
        cls = type(self).__name__
        Rformat = '< {} {} of Type {} with {} reacts and {} products >'.format
        return Rformat(cls,i,'=>',n,m)
    def __eq__(self,other):
        return str(self) == str(other)
    def __contains__(self,item):
        return item in self.compounds

    def coefficient(self,compound):
        """
        Returns the global coefficient of the compound in the reaction if
        the compound is in the reaction, otherwise returns a 0
        This function acts as a wrapper for readability
        """
        return self.compounds.get(compound,0)

    @property
    def reactants_energy(self):
        return sum([c.energy for c in self.reactants.elements()])

    @property
    def AE(self):
        AE = self.TS.energy - self.reactants_energy
        return  AE
    @AE.setter
    def AE(self,other):
        reactants_energy = sum([c.energy for c in self.reactants.elements()])
        try:
            self.TS.energy = reactants_energy + other
        except AttributeError as e:
            raise RuntimeError('Attempting to set the activation energy to '\
                                 'a TSless reaction.')

    @property
    def k(self):
        try:
            k = self.ActE2k(self.T,self.AE)
        except ValueError as e:
            msg = str(e) + ' in reaction:\n\t{}'.format(str(self))
            e.args = (msg,)
            raise e
        return k

    @staticmethod
    def ActE2k(T,AE):
        """
        Transforms an activation energy to a kinetic constant. Uses TST and
        Erying Equation with the transmission coefficient equal to 1. The
        partition coefficients correspond to fluid phase.

        Parameters
        ----------
        T : float
            Temperature in Kelvin.
        AE : Energy
            Activation energy.

        Returns
        -------
        float

        Raises
        -------
        ValueError
            If a negative activation energy is provided

        """
        kb = 1.38064852E-023
        h =  6.62607004E-034
        R =  8.31445985
        if AE < 0:
            raise ValueError('Negative Activation energy found')
        AE_SI  = AE.to_SI()
        k = (kb*(T)/h)*numpy.exp(-AE_SI/(R*(T)))
        return k

class ChemicalSystem(object):
    """
    This class represents a chemical system, understanding
    it as a set of compounds, reactions, barriers and Free Energies
    of Reaction, it includes the utilities for human representation
    (A + B => C) and the matrix representation.

    Parameters
    ----------
    T : float
        Temperature in K (the default is 298.15).
    unit : str
        Default unit to assume when any conflict arises. It is
        also the default unit to report all energies.

    Attributes
    ----------
    shape : tuple
        (n_reactions, n_compounds)
    species : int
        n_compounds
    Name2Compound : dict
        Mapping used for accesing the compounds by name
    Name2TS : dict
        Mapping used for accesing the Transition States by name
    compounds : list
    reactions : list
    transitionstates : list
    T : float
        Temperature in K. Setting this attribute will set the same T for all the
        reactions in the system.

    """
    def __init__(self,T=298.15,unit='kcal/mol'):
        self._T = T # K
        self.compounds = []
        self.reactions = []
        self.transitionstates = []
        self.Name2Compound = dict()
        self.Name2TS = dict()
        self.unit = unit

    def __repr__(self):
        n,m = self.shape
        cls = self.__class__.__name__
        return '<{} with {} compounds and {} reactions>'.format(cls,m,n)
    def __contains__(self,item):
        return item in self.compounds or item in self.reactions

    @property
    def shape(self): 
        return len(self.reactions),len(self.compounds)
    
    @property
    def species(self):
        return len(self.compounds)
    
    @property
    def T(self):
        return self._T
    @T.setter
    def T(self,other):
        self._T = other
        for reaction in self.reactions:
            reaction.T = other

    # Methods related with the (c)ompounds attribute
    def cadd(self,compound,update=False):
        """
        Forcefully adds a compound object and updates the dictionaries
        overwriting it if it already existed.

        Parameters
        ----------
        compound : Compound
        update : bool
            Whether to update the compounds in the system after adding it or not
            (the default is False).
        """
        if compound.label in self.Name2Compound:
            index = self.Name2Compound[compound.label].key
            old = self.compounds[index]
            self.compounds[index] = compound
            raise Warning(f'Overwrote {old} at {self} with {compound}')
        else:
            self.compounds.append(compound)
        self.Name2Compound[compound.label] = compound
        if update:
            self.cupdate()
    def cextend(self,compounds,update=True):
        """
        Lazy Addition, it first adds all the compounds and then
        it updates them.

        Parameters
        ----------
        compounds : list
            list of `Compound` instances
        update : bool
            Whether to update the compounds in the system after adding it or not
            (the default is True).
        """
        for compound in compounds:
            self.cadd(compound,update=False)
        if update:
            self.cupdate()
    def cupdate(self,start=0,update_keys=False):
        """
        Updates the `compounds` attr reinitializing the indices.

        Parameters
        ----------
        start : int
            Positive integrer that represents the initial index used to
            represent compounds. This parameter should only be changed when
            writing programs for non-python languages or for including the T
            and/or V in the differential equations.(the default is 0)
        update_keys : bool
            if True it will update the `Name2Compound` attr as well as the 
            keys of each compound in it. (the default is False).
        Raises
        -------
        AssertionError
            If the start value is negative the whole
        """
        assert start >= 0
        compounds = self.compounds
        for i,C in enumerate(compounds):
            C.key = i + start
            if update_keys:
                self.Name2Compound[C.label] = C

    # Methods related with the (r)eactions attribute
    def radd(self,reaction,update=True):
        """
        Adds and updates a Reaction Instance avoiding duplicates of it within 
        the ChemicalSystem.

        Parameters
        ----------
        reaction : Reaction
            reaction to add to the system.
        update : bool
            if True it modifies the state of the Reaction Instance
            (the default is True).
        """
        if reaction not in self:
            if update:
                if not self.reactions:
                    key = 0 + 1
                else:
                    key = self.reactions[-1].key +1
                reaction.key = key
                reaction.T = self.T
            self.reactions.append(reaction)
        if reaction.TS not in self.transitionstates: 
            self.transitionstates.append(reaction.TS)
    def rextend(self,reactions,update=True):
        """
        Lazy Addition, it first adds all the reactions and then
        it updates them.

        Parameters
        ----------
        reactions : list
            list of Reaction instances
        update : bool
            Whether to update the reactions in the system after adding it or not
            (the default is True).
        """
        for reaction in reactions:
            self.radd(reaction,update=False)
        if update: 
            self.rupdate()
    def rupdate(self,start=0):
        """
        Updates the `reactions` attr ensuring appropiate key and T of each
        Reaction Instance.

        Parameters
        ----------
        start : int
            Initial value for the key in the reactions (the default is 0).
        """
        reactions = self.reactions
        self.transitionstates = []
        for i,R in enumerate(reactions):
            R.key = i + start
            R.T = self.T
            if R.TS not in self.transitionstates: 
                self.transitionstates.append(R.TS)
        self.Name2TS = {ts.label:ts for ts in self.transitionstates}

    # Utilities
    def RxCMatrix(self):
        """
        Returns a matrix that represents the system in which per each
        Matrix[i,j] corresponds to the coefficient of the j compound in the
        i reaction. Possitive if its a product and negative if its a reactant.

        Returns
        -------
        numpy.array

        Raises
        -------
        OperationError
            If the ChemicalSystem is not Updated
        """
        RxCMat = numpy.zeros(shape=self.shape)
        for i,reaction in enumerate(self.reactions):
            for R in reaction.reactants.elements(): # For each Reactant
                RxCMat[i,R.key] -= 1.0
            for P in reaction.products.elements(): # for each Product
                RxCMat[i,P.key] += 1.0
        return RxCMat

    # Main API
    def massbalance(self,compound):
        """
        Creates a MassBalance instance from the mass balance of the compound 
        provided in the chemical system.

        Parameters
        ----------
        compound : Compound
            Compound whose mass balance is to be obtained

        Returns
        -------
        MB
            MassBalance object
        """
        MB = MassBalance(compound)
        n = len(self.reactions)
        for reaction in self.reactions:
            coef = reaction.coefficient(compound)
            # compound in reaction and does not appear as reactant and product
            if abs(coef) > 0:
                MB.items.append((coef,reaction))
        return MB
    def jacobian(self):
        """
        returns a list of Jacobian elements for the mass balances of all the 
        compounds in the system.

        Returns
        -------
        Out
            list of JacobianElement items.
        """
        Out = []
        for compound in self.compounds:
            MB = self.massbalance(compound)
            for compound2 in self.compounds:
                Out.append(MB.partial(compound2))
        return Out


# Parameter clases
class Parameters(UserDict):
    """
    Base class for the Convergence and Simulation Parameters. 
    Provides a construction method from a file. The class should behave like a 
    dictionary.
    """
    @classmethod
    def read_from(cls,file):
        """
        Reads a file line by line and transforms it into a dictionary.

        Parameters
        ----------
        file : str
            A valid filepath to a file.

        Returns
        -------
        dict
            dictionary containing first column as key and an empty string or
            the rest of columns as a string
        """
        parameters = cls()
        with open(file,'r') as F:
            for line in F:
                if line.strip():
                    Aux = line.strip().split(maxsplit=1)
                    key = Aux[0]
                    try:
                        item = Aux[1]
                    except ValueError:
                        item = ''
                    finally:
                        parameters[Aux[0]] = Aux[1]
        return parameters
class SimulationParameters(Parameters):
    """
    Dictionary with default values for simulations. It includes the keys:
    {tfin, dt, trep, concentrations}.
    """
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self['tfin'] = self.get('tfin',0)
        self['trep'] = self.get('trep',0)
        self['dt'] = self.get('dt',0)
        self['concentrations'] = self.get('concentrations',{'index':'value'})
    def read_concentrations(self):
        compounds_mark = ';'
        values_mark  = ','
        text = self.get('concentrations','index')
        self['concentrations'] = dict()
        if text != 'index':
            for compound in text.strip().split(compounds_mark): 
                key,value = compound.split(values_mark,maxsplit=1)
                key_int = int(key.strip())
                if values_mark not in value:
                    self['concentrations'][key_int] = float(value.strip())
                else:
                    values = tuple([float(v.strip()) for v in value.split(',')])
                    self['concentrations'][key_int] = values
    @classmethod
    def read_from(cls,file):
        parameters = super().read_from(file) 
        # The default 'concentrations' has been overridden 
        parameters.read_concentrations()
        return parameters
class ConvergenceParameters(Parameters):
    """
    Dictionary with default values for the parameters needed for the convergence
    of the differential equation solvers to proceed. It includes the keys:
    {rtol, atol}.
    """
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        # default values
        self['rtol'] = self.get('rtol','1E-6')
        self['atol'] = self.get('atol','1E-12')

    @property
    def variables(self):
        return [key for key in self]
    
    def as_str(self,sep=','):
        """ Joins all the key=val pairs by the selected 'sep' string """
        parameters = [f'{key}={val}' for key,val in self.items()] 
        return sep.join(parameters)



#!/usr/bin/python3
import argparse
import subprocess
from itertools import chain

from pathlib import Path

from pykinetic.classes import Energy,SimulationParameters,ConvergenceParameters
from pykinetic.writers import CplusplusWriter
import pykinetic.writers as writers
from pykinetic.utils import (ScannableChemicalSystem, write_indexfile,
                              calc_standard_state_correction)
from pykinetic.userinput import populate_chemicalsystem_fromfiles

__version__ = "0.1.0"

WRITERS = {'python':writers.python.Batch,
           'c++':CplusplusWriter}
COMMANDS = {}
COMMANDS['python'] = lambda x: ['python3',x ]
COMMANDS['c++'] = lambda x: ['g++',str(x),';',f'./{x}']

def create_parser():
    """ Create a Command line argument parser """
    description =""" Creates a python script that contains a
    system of differential equations that represent a Chemical System
    and the tools to solve it numerically"""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("compounds",
                        type=Path,
                        help="File with the compounds and Energies")
    parser.add_argument("reactions",
                        type=Path, 
                        help="File with the reactions, energies and/or TSs")
    parser.add_argument("start",
                        type=float,
                        help="Starting value for the correction scan")
    parser.add_argument("stop",
                        type=float,
                        help="Final value for the correction scan")
    parser.add_argument("steps",type=int,
                        help="Number of steps in between")
    parser.add_argument("--scan-all",
                        dest="scan_all", 
                        action='store_true',
                        default=False,
                        help="""Adds all compounds and non-diffusion TS
                        (TSs associated to reactions of the type <d>) to the 
                        pool of scannable energies if they were not yet included
                        in the corresponding input file with the scannable 
                        keyword""")
    parser.add_argument("--scanunit",
                        default='kcal/mol',
                        choices=Energy._units,
                        help="Unit of Start and Stop parameters (default: kcal/mol)")
    parser.add_argument("--relative",
                        action='store_true',
                        default=False,
                        help="""Whether the energy in the reactions file is 
                        relative to reactants. Otherwise the absolute value of 
                        the TS structure is assumed to correspond to the 
                        provided energy""")
    parser.add_argument("-I","--IndexFile",
                        action="store_true",
                        default=False,
                        help="""If Enabled creates a txt file named 
                        'OutFile'.index with the indices used for reactions,
                        compounds and TSs""")
    parser.add_argument("-Ir","--indexfile-relative",
                        action="store_true",
                        dest="indexfile_relative",
                        default=False,
                        help="""Enforces relative barriers in the specified 
                        default energy unit in the indexfile.""")
    parser.add_argument("-T","--Temperature", 
                        nargs=2,
                        metavar=("Temp","Unit"),
                        default=(25.0,'C'),
                        help="Temperature and unit ('C'=Celsius and 'K'=Kelvin)")
    parser.add_argument("--unit",
                        default='kcal/mol',
                        help="""Energy unit used in all the energy values whose 
                        unit is not specified""", 
                        choices=Energy._units)
    parser.add_argument("--bias",
                        nargs='?',
                        default=0.0,
                        const='Standard State',
                        help="""Applies a bias to all energies. If no argument 
                        is provided it applies Standard State correction. 
                        Otherwise, uses the specified value """)
    parser.add_argument("--convergence", 
                        type=Path,
                        default=None,
                        help="Path to the file with the Convergence parameters")
    parser.add_argument("--simulation",
                        type=Path,
                        default=None,
                        help="Path to the file with the Simulation parameters")
    parser.add_argument("--tail",
                        type=Path,
                        default=None,
                        help="""Path to the non-standard Template for the ending
                         part of the generated script""")
    parser.add_argument("--header",
                        type=Path,
                        default=None,
                        help=""" Path to the non-standard Template for the head 
                        of the generated script""")
    parser.add_argument("--writer",
                        choices=['python','c++'],
                        default='python',
                        help="format for the output model")
    parser.add_argument("--scripts",
                        default=False,
                        action="store_true",
                        help="""If Enabled retains the generated scripts""")
    parser.add_argument("--dryrun",
                        default=False,
                        action="store_true",
                        help="""If enabled it will generate the scripts but wont
                        run them""")
                        
    return parser
def parse_arguments(parser):
    """ Check the arguments of the parser and prepare them for their use """
    args = parser.parse_args()

    # Read the header if it has been specified
    if args.header is not None: 
        with open(args.header) as F:
            txt = F.read()
        args.header = txt
    # Read the tail if it has been specified
    if args.tail is not None: 
        with open(args.tail) as F:
            txt = F.read()
        args.tail = txt

    # Handle Temperature
    if args.Temperature[1] == 'C':
        T = float(args.Temperature[0]) + 273.15
    elif args.Temperature[1] == 'K':
        T = float(args.Temperature[0])
    else:
        msg = "Temperature in {} not implemented".format(args.Temperature[1])
        raise NotImplementedError(msg)
    args.Temperature = T

    # Handle bias
    if args.bias == 'Standard State':
        bias = calc_standard_state_correction(T)
        bias.to_unit(args.unit)
    else:
        bias = Energy(args.bias,args.unit)
    args.bias = bias

    # Handle simulation parameters
    if args.simulation is None: 
        sim_params = SimulationParameters()
    else:
        sim_params = SimulationParameters.read_from(args.simulation)
    args.simulation = sim_params

    # Handle convergence parameters
    if args.convergence is None:
        conv_params = ConvergenceParameters()
    else:
        conv_params = ConvergenceParameters.read_from(args.convergence)
    args.convergence = conv_params

    if args.dryrun:
        args.scripts = True
    
    if not args.IndexFile and args.indexfile_relative: 
        args.IndexFile = True

    return args

def main():
    parser = create_parser()
    args = parse_arguments(parser)
    unit = args.unit
    scan_unit = args.scanunit
    # Prepare Filesystem
    outputs_folder = Path('Outputs')
    indices_folder = Path('IndexFiles')
    scripts_folder = Path('ScriptFiles')
    scan_parameters_file = Path('Scan_Params.out.txt')
    tmp_name = 'Temp'
    if args.writer == 'python':
        suffix = 'py'
    elif args.writer == 'c++':
        suffix = 'cpp'
    tmp_file = Path(f'{tmp_name}.{suffix}')
    if not args.dryrun:
        outputs_folder.mkdir(exist_ok=True)
    if args.IndexFile:
        indices_folder.mkdir(exist_ok=True)
    if args.scripts:
        scripts_folder.mkdir(exist_ok=True)

    # Initialize the Writer
    writer_cls = WRITERS[args.writer]
    writer = writer_cls(conc_var='x',mb_var='dxdt',fun_var='model',
                        jac_var='Jac',jac_fun_var='jacobian',
                        header=args.header,tail=args.tail)
    writer.set_parameters(simulation=args.simulation,
                          convergence=args.convergence)

    # Initialize the ChemicalSystem
    step = (args.stop - args.start)/args.steps
    steps = args.steps + 1 # in order to include the final point.
    energies = [Energy(args.start + step*i,args.scanunit) for i in range(steps)]

    # Record in a file the value of the correction of each step
    int_f = '{:03d}'.format
    energy_f = lambda x: str(x).split()[0]
    with open(scan_parameters_file,'w') as F:
        F.write(f'Scan Step    Energy({args.scanunit})\n')
    
    # Initialize the ChemicalSystem 
    chemsys = ScannableChemicalSystem(scan_unit=scan_unit,unit=unit,
                                      bias=args.bias,T=args.Temperature)
    populate_chemicalsystem_fromfiles(chemsys,
                                      file_c=args.compounds,
                                      file_r=args.reactions,
                                      energy_unit=unit,
                                      relativeE=args.relative)
    if args.scan_all:
        for compound in chemsys.compounds: 
            compound.scannable = True
        for ts in chemsys.transitionstates: 
            if not hasattr(ts,'barrier'): # only DiffusionTS have .barrier
                ts.scannable = True
    chemsys.apply_bias()
    chemsys.apply_scan()

    # Ensure that at least one compound or TS has the scannable attribute set up
    for item in chain(chemsys.compounds,chemsys.transitionstates): 
        if item.scannable:
            break
    else:
        raise ValueError('''No Compound/TS has the "scan" keyword in the 
        input files. Please use pykinetic-model.py if the input files are 
        correct otherwise please indicate in the input files which Compounds 
        and/or transition states should be scanned.''')

    for i,energy in enumerate(energies): # Keep going
        stem = f'scan_{i:03d}' # stem of the produced files
        chemsys.scan = energy

        # Create Name of Script and OutFile
        out_data_file = outputs_folder/f'{stem}.data'
        writer.parameters['out_filename'] = out_data_file
        print(stem)

        # Write the script
        writer.write(chemsys,tmp_file)

        if not args.dryrun:
            # Run Script as subprocess
            cmd = COMMANDS[args.writer](tmp_file)
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            #for line in p.stdout:
            #    print(line)
            # Wait for it to finish
            p.wait()

        if args.IndexFile:
            index_file = indices_folder / f'{stem}.index'
            isrelative = args.indexfile_relative or args.relative
            write_indexfile(chemsys,index_file, isrelative=isrelative)
        
        if args.scripts:
            script = scripts_folder / f'{stem}{tmp_file.suffix}'
            tmp_file.rename(script)

        with open(scan_parameters_file,'a') as F:
            F.write('{: ^9}\t{: ^16}\n'.format(f'{i:03d}',str(energy)))

if __name__ == "__main__":
    main()



import unittest
from pykinetic.writers._base import TEMPLATES_PATH
from pykinetic.classes import Compound, Energy, ChemicalSystem, Reaction, TransitionState
from pykinetic.writers import Indent, CplusplusWriter
from pykinetic.writers.python import Batch as pyBatch
from pykinetic.writers.python import SemiBatch as pySemiBatch
from pykinetic.writers.python import SemiBatchExtended as pySemiBatchExtended
from pykinetic.writers.python import PFR as pyPFR

class IndentTest(unittest.TestCase):
    def test_default(self):
        items =  [(['Test',]        ,['    Test',]),
                  (['Test1','Test2'],['    Test1','    Test2']),
                  ( 'Test'          ,['    T','    e','    s','    t']),
                  (   ''            ,[]),
                  (   []            ,[]),]
        for t,s in items:
            with self.subTest(t=t):
                self.assertEqual(Indent(t),s)
    def test_level(self):
        items =  [(1, ['Test',]        ,['\tTest',]),
                  (1, ['Test1','Test2'],['\tTest1','\tTest2']),
                  (2, ['Test1','Test2'],['\t\tTest1','\t\tTest2']),
                  (2,  'Test'          ,['\t\tT','\t\te','\t\ts','\t\tt']),
                  (4,    []            ,[]),]
        for i,t,s in items:
            with self.subTest(t=t):
                self.assertEqual(Indent(t,'\t',level=i),s)
    def test_level_consistency(self):
        test = 'line0 line1 line2 line3 line4 line-1'.split()
        for i,j in zip(range(4),range(1,5)):
            s = Indent(test,level=j)
            t = Indent(Indent(test,level=i),level=1)
            with self.subTest(i=i,j=j):
                self.assertEqual(t,s)

class pyBatchTest(unittest.TestCase):
    def setUp(self):
        self.writer = pyBatch()
        unit = 'kcal/mol'
        A = Compound('A',Energy( 0.0,unit))
        B = Compound('B',Energy( 0.0,unit))
        C = Compound('C',Energy( 2.0,unit))
        D = Compound('D',Energy(-1.0,unit))
        E = Compound('E',Energy(  99,unit)) # Does not appear in reactions
        # TransitionState(Energy(1.0,unit))
        self.compounds = [A,B,C,D,E]
        self.chemsys = ChemicalSystem()
        for c in self.compounds:
            self.chemsys.cadd(c,update=False)
        self.chemsys.cupdate()
        self.reactions = [Reaction((A,),(C,)),
                          Reaction((A,B),(C,)),
                          Reaction((A,A),(C,)),
                          Reaction((A,),(B,C)),
                          Reaction((A,),(C,B)),
                          Reaction((A,),(C,C)),
                          Reaction((A,B),(C,D)),
                          Reaction((A,B,C),(D,)),
                          Reaction((A,),(B,C,D))]
        for r in self.reactions:
            r.TS = TransitionState(Energy(1.0,'kcal/mol') + r.reactants_energy)
            self.chemsys.radd(r,update=False)
        self.chemsys.rupdate()

    def test_templateload(self):
        with open(TEMPLATES_PATH/'python_batch.head') as F:
            sol_h = F.read()
        with open(TEMPLATES_PATH/'python_batch.tail') as F:
            sol_t = F.read()
        with self.subTest(template='head'):
            self.assertEqual(sol_h,self.writer._header)
        with self.subTest(template='tail'):
            self.assertEqual(sol_t,self.writer._tail)
    def test_ratelaw_expr(self):
        solutions = ['k00*x[0]',
                     'k01*x[0]*x[1]',
                     'k02*x[0]*x[0]',
                     'k03*x[0]',
                     'k04*x[0]',
                     'k05*x[0]',
                     'k06*x[0]*x[1]',
                     'k07*x[0]*x[1]*x[2]',
                     'k08*x[0]']
        for reaction,solution in zip(self.reactions,solutions):
            with self.subTest(reaction=reaction):
                _,test = self.writer.ratelaw(reaction)
                self.assertEqual(test,solution)
    def test_ratelaw_var(self):
        solutions = ['r00','r01','r02','r03','r04','r05','r06','r07','r08']
        for reaction,solution in zip(self.reactions,solutions):
            with self.subTest(reaction=reaction):
                test,_  = self.writer.ratelaw(reaction)
                self.assertEqual(test,solution)
    def test_ratelaw_partial(self):
        solutions = ['k00',
                     'k01*x[1]',
                     '2.0*k02*x[0]',
                     'k03',
                     'k04',
                     'k05',
                     'k06*x[1]',
                     'k07*x[1]*x[2]',
                     '',
                     '']
        compounds = [self.compounds[0] for i in solutions[:-2]]
        compounds.append(self.compounds[-2]) # Add Product
        compounds.append(self.compounds[-1]) # Add external
        for r,s,c in zip(self.reactions,solutions,compounds):
            with self.subTest(reaction=str(r),compound=c):
                test = self.writer.ratelaw_partial(r,c)
                self.assertEqual(test,s)
    def test_massbalance_expr(self):
        solutions = ['-r00-r01-2.0*r02-r03-r04-r05-r06-r07-r08',
                     '-r01+r03+r04-r06-r07+r08',
                     '+r00+r01+r02+r03+r04+2.0*r05+r06-r07+r08',
                     '+r06+r07+r08',
                     '0',
                     '0']
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for s,c in zip(solutions,compounds):
            with self.subTest(compound=c):
                ## Future MB = self.chemsys.massbalance(c)
                MB = self.chemsys.massbalance(c)
                _ , test = self.writer.massbalance(MB)
                self.assertEqual(test,s)
    def test_massbalance_var(self):
        solutions = ['dxdt[0]',
                     'dxdt[1]',
                     'dxdt[2]',
                     'dxdt[3]',
                     '',
                     '']
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for s,c in zip(solutions,compounds):
            with self.subTest(compound=c):
                ## Future MB = self.chemsys.massbalance(c)
                MB = self.chemsys.massbalance(c)
                test, _ = self.writer.massbalance(MB)
                self.assertEqual(test,s)
    def test_jacobian_element_expr(self):
        solutions = [['-k00-k01*x[1]-2.0*2.0*k02*x[0]-k03-k04-k05-k06*x[1]-k07*x[1]*x[2]-k08',
                      '-k01*x[0]-k06*x[0]-k07*x[0]*x[2]',
                      '-k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['-k01*x[1]+k03+k04-k06*x[1]-k07*x[1]*x[2]+k08',
                      '-k01*x[0]-k06*x[0]-k07*x[0]*x[2]',
                      '-k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['+k00+k01*x[1]+2.0*k02*x[0]+k03+k04+2.0*k05+k06*x[1]-k07*x[1]*x[2]+k08',
                      '+k01*x[0]+k06*x[0]-k07*x[0]*x[2]',
                      '-k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['+k06*x[1]+k07*x[1]*x[2]+k08',
                      '+k06*x[0]+k07*x[0]*x[2]',
                      '+k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['0']*6]
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for sols,c in zip(solutions,compounds):
            ## Future MB = self.chemsys.massbalance(c)
            MB = self.chemsys.massbalance(c)
            ## Future Jacobian
            for c2,s in zip(compounds,sols):
                Jac_ij = MB.partial(c2)
                with self.subTest(compound1=c,compound2=c2):
                    _, test = self.writer.jacobian_element(Jac_ij)
                    self.assertEqual(test,s)
    def test_jacobian_element_var(self):
        solutions = [['Jac[0,0]',
                      'Jac[0,1]',
                      'Jac[0,2]',
                      'Jac[0,3]',
                      '',
                      ''],
                     ['Jac[1,0]',
                      'Jac[1,1]',
                      'Jac[1,2]',
                      'Jac[1,3]',
                      '',
                      ''],
                     ['Jac[2,0]',
                      'Jac[2,1]',
                      'Jac[2,2]',
                      'Jac[2,3]',
                      '',
                      ''],
                     ['Jac[3,0]',
                      'Jac[3,1]',
                      'Jac[3,2]',
                      'Jac[3,3]',
                      '',
                      ''],
                     ['']*6,
                     ['']*6]
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for sols,c in zip(solutions,compounds):
            ## Future MB = self.chemsys.massbalance(c)
            MB = self.chemsys.massbalance(c)
            ## Future Jacobian
            for c2,s in zip(compounds,sols):
                Jac_ij = MB.partial(c2)
                with self.subTest(compound1=c,compound2=c2):
                    test, _ = self.writer.jacobian_element(Jac_ij)
                    self.assertEqual(test,s)

    # test "private" methods
    def test__kinetic_constants(self):
        solution = ['#Constants',
                    'k00 = 1.1488119097e+12',
                    'k01 = 1.1488119097e+12',
                    'k02 = 1.1488119097e+12',
                    'k03 = 1.1488119097e+12',
                    'k04 = 1.1488119097e+12',
                    'k05 = 1.1488119097e+12',
                    'k06 = 1.1488119097e+12',
                    'k07 = 1.1488119097e+12',
                    'k08 = 1.1488119097e+12']
        test = self.writer._kinetic_constants(self.chemsys)
        self.assertEqual(test,solution)
    def test__ratelaws(self):
        solution = ['#Ratelaws',
                    'r00 = k00*x[0]',
                    'r01 = k01*x[0]*x[1]',
                    'r02 = k02*x[0]*x[0]',
                    'r03 = k03*x[0]',
                    'r04 = k04*x[0]',
                    'r05 = k05*x[0]',
                    'r06 = k06*x[0]*x[1]',
                    'r07 = k07*x[0]*x[1]*x[2]',
                    'r08 = k08*x[0]']
        test = self.writer._ratelaws(self.chemsys)
        self.assertEqual(test,solution)
    def test__massbalances(self):
        solution = ['#MassBalances',
                    'dxdt[0] = -r00-r01-2.0*r02-r03-r04-r05-r06-r07-r08',
                    'dxdt[1] = -r01+r03+r04-r06-r07+r08',
                    'dxdt[2] = +r00+r01+r02+r03+r04+2.0*r05+r06-r07+r08',
                    'dxdt[3] = +r06+r07+r08']
        test = self.writer._massbalances(self.chemsys)
        self.assertEqual(test,solution)
    def test__jacobian_elements(self):
        solution = ['#Non-zero Elements',
                    'Jac[0,0] = -k00-k01*x[1]-2.0*2.0*k02*x[0]-k03-k04-k05-k06*x[1]-k07*x[1]*x[2]-k08',
                    'Jac[0,1] = -k01*x[0]-k06*x[0]-k07*x[0]*x[2]',
                    'Jac[0,2] = -k07*x[0]*x[1]',
                    'Jac[1,0] = -k01*x[1]+k03+k04-k06*x[1]-k07*x[1]*x[2]+k08',
                    'Jac[1,1] = -k01*x[0]-k06*x[0]-k07*x[0]*x[2]',
                    'Jac[1,2] = -k07*x[0]*x[1]',
                    'Jac[2,0] = +k00+k01*x[1]+2.0*k02*x[0]+k03+k04+2.0*k05+k06*x[1]-k07*x[1]*x[2]+k08',
                    'Jac[2,1] = +k01*x[0]+k06*x[0]-k07*x[0]*x[2]',
                    'Jac[2,2] = -k07*x[0]*x[1]',
                    'Jac[3,0] = +k06*x[1]+k07*x[1]*x[2]+k08',
                    'Jac[3,1] = +k06*x[0]+k07*x[0]*x[2]',
                    'Jac[3,2] = +k07*x[0]*x[1]']
        test = self.writer._jacobian_elements(self.chemsys)
        self.assertEqual(test,solution)
class PySemiBatchTest(pyBatchTest):
    def setUp(self):
        self.concentrations = {0:(0.0,1.0),1:(1.0,0.0)}
        self.flow = 1.0
        self.Vini = 1.0
        self.writer = pySemiBatch(flow=self.flow,Vini=self.Vini)
        unit = 'kcal/mol'
        A = Compound('A',Energy( 0.0,unit))
        B = Compound('B',Energy( 0.0,unit))
        C = Compound('C',Energy( 2.0,unit))
        D = Compound('D',Energy(-1.0,unit))
        E = Compound('E',Energy(  99,unit)) # Does not appear in reactions
        # TransitionState(Energy(1.0,unit))
        self.compounds = [A,B,C,D,E]
        self.chemsys = ChemicalSystem()
        for c in self.compounds:
            self.chemsys.cadd(c,update=False)
        self.chemsys.cupdate()
        self.reactions = [Reaction((A,),(C,)),
                          Reaction((A,B),(C,)),
                          Reaction((A,A),(C,)),
                          Reaction((A,),(B,C)),
                          Reaction((A,),(C,B)),
                          Reaction((A,),(C,C)),
                          Reaction((A,B),(C,D)),
                          Reaction((A,B,C),(D,)),
                          Reaction((A,),(B,C,D))]
        for r in self.reactions:
            r.TS = TransitionState(Energy(1.0,'kcal/mol') + r.reactants_energy)
            self.chemsys.radd(r,update=False)
        self.chemsys.rupdate()
    def test_templateload(self):
        with open(TEMPLATES_PATH/'python_semibatch.head') as F:
            sol_h = F.read()
        with open(TEMPLATES_PATH/'python_semibatch.tail') as F:
            sol_t = F.read()
        with self.subTest(template='head'):
            self.assertEqual(sol_h,self.writer._header)
        with self.subTest(template='tail'):
            self.assertEqual(sol_t,self.writer._tail)
    def test__set_initial_concentrations(self):
        self.writer._set_concentrations(self.concentrations)
        self.assertEqual(self.writer.keys,[0,])
        self.assertDictEqual(self.writer.Cadd,{0:1.0})
        self.assertDictEqual(self.writer.parameters['concentrations'],{1:1.0})
    def test__flask_initial_concentrations(self):
        sol = f'Cin{0:02.0f} = {self.concentrations[0][1]}'
        self.writer._set_concentrations(self.concentrations)
        test = self.writer._flask_initial_concentrations()
        self.assertEqual([sol,],test)
    def test_massbalance_expr(self):
        self.writer._set_concentrations(self.concentrations)
        solutions = ['tflow*Cin00 -r00-r01-2.0*r02-r03-r04-r05-r06-r07-r08',
                     '-r01+r03+r04-r06-r07+r08',
                     '+r00+r01+r02+r03+r04+2.0*r05+r06-r07+r08',
                     '+r06+r07+r08',
                     '0',
                     '0']
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for s,c in zip(solutions,compounds):
            with self.subTest(compound=c):
                ## Future MB = self.chemsys.massbalance(c)
                MB = self.chemsys.massbalance(c)
                _ , test = self.writer.massbalance(MB)
                self.assertEqual(test,s)
    def test_massbalance_var(self):
        self.writer._set_concentrations(self.concentrations)
        solutions = ['dxdt[0]',
                     'dxdt[1]',
                     'dxdt[2]',
                     'dxdt[3]',
                     '',
                     '']
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for s,c in zip(solutions,compounds):
            with self.subTest(compound=c):
                ## Future MB = self.chemsys.massbalance(c)
                MB = self.chemsys.massbalance(c)
                test, _ = self.writer.massbalance(MB)
                self.assertEqual(test,s)
class pySemiBatchExtendedTest(pyBatchTest): 
    def setUp(self):
        self.concentrations = {0:(0.0,1.0),1:(1.0,0.0)}
        self.flow = 1.0
        self.Vini = 1.0
        self.writer = pySemiBatchExtended(flow=self.flow,Vini=self.Vini)
        unit = 'kcal/mol'
        A = Compound('A',Energy( 0.0,unit))
        B = Compound('B',Energy( 0.0,unit))
        C = Compound('C',Energy( 2.0,unit))
        D = Compound('D',Energy(-1.0,unit))
        E = Compound('E',Energy(  99,unit)) # Does not appear in reactions
        # TransitionState(Energy(1.0,unit))
        self.compounds = [A,B,C,D,E]
        self.chemsys = ChemicalSystem()
        for c in self.compounds:
            self.chemsys.cadd(c,update=False)
        self.chemsys.cupdate()
        self.reactions = [Reaction((A,),(C,)),
                          Reaction((A,B),(C,)),
                          Reaction((A,A),(C,)),
                          Reaction((A,),(B,C)),
                          Reaction((A,),(C,B)),
                          Reaction((A,),(C,C)),
                          Reaction((A,B),(C,D)),
                          Reaction((A,B,C),(D,)),
                          Reaction((A,),(B,C,D))]
        for r in self.reactions:
            r.TS = TransitionState(Energy(1.0,'kcal/mol') + r.reactants_energy)
            self.chemsys.radd(r,update=False)
        self.chemsys.rupdate()
    def test_templateload(self):
        with open(TEMPLATES_PATH/'python_semibatch_extended.head') as F:
            sol_h = F.read()
        with open(TEMPLATES_PATH/'python_semibatch_extended.tail') as F:
            sol_t = F.read()
        with self.subTest(template='head'):
            self.assertEqual(sol_h,self.writer._header)
        with self.subTest(template='tail'):
            self.assertEqual(sol_t,self.writer._tail)
    

class PyPFRTest(pyBatchTest): 
    def setUp(self):
        self.concentrations = {0:(0.0,1.0),1:(1.0,0.0)}
        self.flow = 1.0
        self.volume = 1.0
        self.writer = pyPFR(flow=self.flow,volume=self.volume)
        unit = 'kcal/mol'
        A = Compound('A',Energy( 0.0,unit))
        B = Compound('B',Energy( 0.0,unit))
        C = Compound('C',Energy( 2.0,unit))
        D = Compound('D',Energy(-1.0,unit))
        E = Compound('E',Energy(  99,unit)) # Does not appear in reactions
        # TransitionState(Energy(1.0,unit))
        self.compounds = [A,B,C,D,E]
        self.chemsys = ChemicalSystem()
        for c in self.compounds:
            self.chemsys.cadd(c,update=False)
        self.chemsys.cupdate()
        self.reactions = [Reaction((A,),(C,)),
                          Reaction((A,B),(C,)),
                          Reaction((A,A),(C,)),
                          Reaction((A,),(B,C)),
                          Reaction((A,),(C,B)),
                          Reaction((A,),(C,C)),
                          Reaction((A,B),(C,D)),
                          Reaction((A,B,C),(D,)),
                          Reaction((A,),(B,C,D))]
        for r in self.reactions:
            r.TS = TransitionState(Energy(1.0,'kcal/mol') + r.reactants_energy)
            self.chemsys.radd(r,update=False)
        self.chemsys.rupdate()
    def test_templateload(self):
        with open(TEMPLATES_PATH/'python_PFR.head') as F:
            sol_h = F.read()
        with open(TEMPLATES_PATH/'python_PFR.tail') as F:
            sol_t = F.read()
        with self.subTest(template='head'):
            self.assertEqual(sol_h,self.writer._header)
        with self.subTest(template='tail'):
            self.assertEqual(sol_t,self.writer._tail)

class CplusplusWriterTest(unittest.TestCase):
    def setUp(self):
        self.writer = CplusplusWriter()
        unit = 'kcal/mol'
        A = Compound('A',Energy( 0.0,unit))
        B = Compound('B',Energy( 0.0,unit))
        C = Compound('C',Energy( 2.0,unit))
        D = Compound('D',Energy(-1.0,unit))
        E = Compound('E',Energy(  99,unit)) # Does not appear in reactions
        self.compounds = [A,B,C,D,E]
        self.chemsys = ChemicalSystem()
        for c in self.compounds:
            self.chemsys.cadd(c,update=False)
        self.chemsys.cupdate()
        self.reactions = [Reaction((A,),(C,)),
                          Reaction((A,B),(C,)),
                          Reaction((A,A),(C,)),
                          Reaction((A,),(B,C)),
                          Reaction((A,),(C,B)),
                          Reaction((A,),(C,C)),
                          Reaction((A,B),(C,D)),
                          Reaction((A,B,C),(D,)),
                          Reaction((A,),(B,C,D))]
        for r in self.reactions:
            r.TS = TransitionState(Energy(1.0,'kcal/mol') + r.reactants_energy)
            self.chemsys.radd(r,update=False)
        self.chemsys.rupdate()
    def test_templateload(self):
        with open(TEMPLATES_PATH/'cplusplus_header.default') as F:
            sol_h = F.read()
        with open(TEMPLATES_PATH/'cplusplus_tail.default') as F:
            sol_t = F.read()
        with self.subTest(template='head'):
            self.assertEqual(sol_h,self.writer._header)
        with self.subTest(template='tail'):
            self.assertEqual(sol_t,self.writer._tail)
    def test_ratelaw_expr(self):
        solutions = ['k00*x[0]',
                     'k01*x[0]*x[1]',
                     'k02*x[0]*x[0]',
                     'k03*x[0]',
                     'k04*x[0]',
                     'k05*x[0]',
                     'k06*x[0]*x[1]',
                     'k07*x[0]*x[1]*x[2]',
                     'k08*x[0]']
        for reaction,solution in zip(self.reactions,solutions):
            with self.subTest(reaction=reaction):
                _,test = self.writer.ratelaw(reaction)
                self.assertEqual(test,solution)
    def test_ratelaw_var(self):
        solutions = ['r00','r01','r02','r03','r04','r05','r06','r07','r08']
        for reaction,solution in zip(self.reactions,solutions):
            with self.subTest(reaction=reaction):
                test,_  = self.writer.ratelaw(reaction)
                self.assertEqual(test,solution)
    def test_ratelaw_partial(self):
        solutions = ['k00',
                     'k01*x[1]',
                     '2.0*k02*x[0]',
                     'k03',
                     'k04',
                     'k05',
                     'k06*x[1]',
                     'k07*x[1]*x[2]',
                     '',
                     '']
        compounds = [self.compounds[0] for i in solutions[:-2]]
        compounds.append(self.compounds[-2]) # Add Product
        compounds.append(self.compounds[-1]) # Add external
        for r,s,c in zip(self.reactions,solutions,compounds):
            with self.subTest(reaction=str(r),compound=c):
                test = self.writer.ratelaw_partial(r,c)
                self.assertEqual(test,s)
    def test_massbalance_expr(self):
        solutions = ['-r00-r01-2.0*r02-r03-r04-r05-r06-r07-r08',
                     '-r01+r03+r04-r06-r07+r08',
                     '+r00+r01+r02+r03+r04+2.0*r05+r06-r07+r08',
                     '+r06+r07+r08',
                     '0',
                     '0']
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for s,c in zip(solutions,compounds):
            with self.subTest(compound=c):
                ## Future MB = self.chemsys.massbalance(c)
                MB = self.chemsys.massbalance(c)
                _ , test = self.writer.massbalance(MB)
                self.assertEqual(test,s)
    def test_massbalance_var(self):
        solutions = ['dxdt[0]',
                     'dxdt[1]',
                     'dxdt[2]',
                     'dxdt[3]',
                     'dxdt[4]',
                     '']
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for s,c in zip(solutions,compounds):
            with self.subTest(compound=c):
                ## Future MB = self.chemsys.massbalance(c)
                MB = self.chemsys.massbalance(c)
                test, _ = self.writer.massbalance(MB)
                self.assertEqual(test,s)
    def test_jacobian_element_expr(self):
        solutions = [['-k00-k01*x[1]-2.0*2.0*k02*x[0]-k03-k04-k05-k06*x[1]-k07*x[1]*x[2]-k08',
                      '-k01*x[0]-k06*x[0]-k07*x[0]*x[2]',
                      '-k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['-k01*x[1]+k03+k04-k06*x[1]-k07*x[1]*x[2]+k08',
                      '-k01*x[0]-k06*x[0]-k07*x[0]*x[2]',
                      '-k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['+k00+k01*x[1]+2.0*k02*x[0]+k03+k04+2.0*k05+k06*x[1]-k07*x[1]*x[2]+k08',
                      '+k01*x[0]+k06*x[0]-k07*x[0]*x[2]',
                      '-k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['+k06*x[1]+k07*x[1]*x[2]+k08',
                      '+k06*x[0]+k07*x[0]*x[2]',
                      '+k07*x[0]*x[1]',
                      '0',
                      '0',
                      '0'],
                     ['0']*6]
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for sols,c in zip(solutions,compounds):
            ## Future MB = self.chemsys.massbalance(c)
            MB = self.chemsys.massbalance(c)
            ## Future Jacobian
            for c2,s in zip(compounds,sols):
                Jac_ij = MB.partial(c2)
                with self.subTest(compound1=c,compound2=c2):
                    _, test = self.writer.jacobian_element(Jac_ij)
                    self.assertEqual(test,s)
    def test_jacobian_element_var(self):
        solutions = [['Jac(0,0)',
                      'Jac(0,1)',
                      'Jac(0,2)',
                      'Jac(0,3)',
                      'Jac(0,4)',
                      ''],
                     ['Jac(1,0)',
                      'Jac(1,1)',
                      'Jac(1,2)',
                      'Jac(1,3)',
                      'Jac(1,4)',
                      ''],
                     ['Jac(2,0)',
                      'Jac(2,1)',
                      'Jac(2,2)',
                      'Jac(2,3)',
                      'Jac(2,4)',
                      ''],
                     ['Jac(3,0)',
                      'Jac(3,1)',
                      'Jac(3,2)',
                      'Jac(3,3)',
                      'Jac(3,4)',
                      ''],
                     ['Jac(4,0)',
                      'Jac(4,1)',
                      'Jac(4,2)',
                      'Jac(4,3)',
                      'Jac(4,4)',
                      ''],
                     ['']*6]
        compounds = [c for c in self.compounds]
        compounds.append(Compound('ClearlyNotInTheSystem',5.2))
        for sols,c in zip(solutions,compounds):
            ## Future MB = self.chemsys.massbalance(c)
            MB = self.chemsys.massbalance(c)
            ## Future Jacobian
            for c2,s in zip(compounds,sols):
                Jac_ij = MB.partial(c2)
                with self.subTest(compound1=c,compound2=c2):
                    test, _ = self.writer.jacobian_element(Jac_ij)
                    self.assertEqual(test,s)

    # test "private" methods
    def test__kinetic_constants(self):
        solution = ['// Kinetic Constants',
                    'const double k00 = 1.1488119097e+12;',
                    'const double k01 = 1.1488119097e+12;',
                    'const double k02 = 1.1488119097e+12;',
                    'const double k03 = 1.1488119097e+12;',
                    'const double k04 = 1.1488119097e+12;',
                    'const double k05 = 1.1488119097e+12;',
                    'const double k06 = 1.1488119097e+12;',
                    'const double k07 = 1.1488119097e+12;',
                    'const double k08 = 1.1488119097e+12;']
        test = self.writer._kinetic_constants(self.chemsys)
        self.assertEqual(test,solution)
    def test__ratelaws(self):
        solution = ['// Ratelaws',
                    'double r00,r01,r02,r03,r04,r05,r06,r07,r08;',
                    'r00 = k00*x[0];',
                    'r01 = k01*x[0]*x[1];',
                    'r02 = k02*x[0]*x[0];',
                    'r03 = k03*x[0];',
                    'r04 = k04*x[0];',
                    'r05 = k05*x[0];',
                    'r06 = k06*x[0]*x[1];',
                    'r07 = k07*x[0]*x[1]*x[2];',
                    'r08 = k08*x[0];']
        test = self.writer._ratelaws(self.chemsys)
        self.assertEqual(test,solution)
    def test__massbalances(self):
        solution = ['// MassBalances',
                    'dxdt[0] = -r00-r01-2.0*r02-r03-r04-r05-r06-r07-r08;',
                    'dxdt[1] = -r01+r03+r04-r06-r07+r08;',
                    'dxdt[2] = +r00+r01+r02+r03+r04+2.0*r05+r06-r07+r08;',
                    'dxdt[3] = +r06+r07+r08;',
                    'dxdt[4] = 0;']
        test = self.writer._massbalances(self.chemsys)
        self.assertEqual(test,solution)



from setuptools import setup, find_packages

__version__ = "0.1.0"

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name = 'pykinetic',
    packages = find_packages(),#['pykinetic',],
    description = """A python library and command line apps to write 
                     microkinetic models for chemistry""",
    keywords = ['compchem', 'microkinetics'],
    author = 'Raúl Pérez-Soto',
    classifiers = ["Programming Language :: Python :: 3",],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url = 'https://github.com/rperezsoto/pykinetic', # Update
    python_requires='>=3.6',
    install_requires=['setuptools','pathlib','numpy'],
    extras_requires=['numpy','scipy'],
    include_package_data=True,
    package_data = {'templates': ['pykinetic/templates/*'],
                    'tests'    : ['pykinetic/tests/*.py']},
    scripts = ['pykinetic/scripts/pykinetic-model.py',
               'pykinetic/scripts/pykinetic-scan.py'],
)



"""
This module contains all the different writers per reactor for the python language.
"""

from collections import Counter
from ._base import TEMPLATES_PATH, Indent, Writer

class Batch(Writer):
    """
    Writer class of python scripts for Batch reactor models
    
    """

    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        for var in ['x', 'dxdt', 'f', 'jac','jac_f']:
            self.parameters[var] = getattr(self,var)
        self.parameters['method'] = 'LSODA'

    def _load_default_header(self):
        with open(TEMPLATES_PATH / 'python_batch.head','r') as F:
            txt = F.read()
        self._header = txt
    def _load_default_tail(self):
        with open(TEMPLATES_PATH / 'python_batch.tail','r') as F:
            txt = F.read()
        self._tail = txt

    def set_parameters(self,simulation=None,convergence=None):
        if simulation is not None:
            for key in ['tfin','trep','dt']: 
                self.parameters[key] = simulation.get(key,'')
            concentrations = self._initial_concentrations(simulation)
            self.parameters['concentrations'] = concentrations
        if convergence is not None:
            self.parameters['convergence'] = convergence.as_str(sep=',')

    # methods for object -> str transformations
    def ratelaw(self,reaction): # Currently only for Elemental Steps
        var = f'r{reaction.key:02.0f}'
        Aux = []
        expr = f'k{reaction.key:02.0f}*'
        reactants = reaction.reactants.elements() # Unpack the reactants Counter
        for R in sorted(reactants,key=lambda x: x.key):
            Aux.append(f'{self.x}[{R.key}]')
        expr += '*'.join(Aux)
        return  var, expr
    def ratelaw_partial(self,reaction,compound): # Currently only for Elemental Steps
        #var = f'r{reaction.key:02.0f}' Legacy code, to remove
        if compound not in reaction.reactants:
            return ''
        Aux = []
        # Create a new Unpack the reactants Counter
        reactants = Counter(reaction.reactants.elements())
        coef = reactants[compound]
        if coef > 1:
            Aux.append(f'{coef:0.1f}*k{reaction.key:02.0f}')
        else:
            Aux.append(f'k{reaction.key:02.0f}')
        reactants[compound] -= 1
        reactants = reactants.elements()
        for R in sorted(reactants,key=lambda x: x.key):
            Aux.append(f'{self.x}[{R.key}]')
        expr = '*'.join(Aux)
        return  expr
    def constant(self,reaction,value_format='{:0.10e}'):  # Currently only for Elemental Steps
        return super().constant(reaction,value_format)
    def massbalance(self,MassBalance):
        if not MassBalance.items:
            return '', '0'
        var = f'{self.dxdt}[{MassBalance.compound.key}]'
        expr = []
        for coef,reaction in MassBalance.items:
            if coef == 1:
                Aux = '+'
            elif coef == -1:
                Aux = '-'
            else:
                Aux = f'{coef:+0.1f}*'
            r_expr = f'{Aux}r{reaction.key:02.0f}'
            expr.append(r_expr)
        return var, ''.join(expr)
    def jacobian_element(self,Jac_ij):
        C1,C2 = Jac_ij.compound1, Jac_ij.compound2
        if not Jac_ij.items:
            return '', '0'
        var = f'{self.jac}[{C1.key},{C2.key}]'
        expr = []
        for coef,reaction in Jac_ij.items:
            if coef == 1:
                Aux = '+'
            elif coef == -1:
                Aux = '-'
            else:
                Aux = f'{coef:+0.1f}*'
            r_expr = self.ratelaw_partial(reaction,C2)
            if not r_expr:
                continue
            expr.append(Aux + r_expr)
        if not expr:
            return var ,'0'
        return var, ''.join(expr)

    # auxiliary writing methods
    def _kinetic_constants(self,chemicalsys):
        constants = ['#Constants',]
        for reaction in chemicalsys.reactions:
            # vars,values = # In case we consider supporting reversible elemental steps?
            var,value = self.constant(reaction)
            constants.append(f'{var} = {value}')
        return constants
    def _ratelaws(self,chemicalsys):
        ratelaws = ['#Ratelaws',]
        for reaction in chemicalsys.reactions:
            var,law = self.ratelaw(reaction)
            ratelaws.append(f'{var} = {law}')
        return ratelaws
    def _massbalances(self,chemicalsys):
        MBs = ['#MassBalances',]
        for compound in chemicalsys.compounds:
            MB = chemicalsys.massbalance(compound)
            var,law = self.massbalance(MB)
            if var:
                MBs.append(f'{var} = {law}')
        return MBs
    def _jacobian_elements(self,chemicalsys):
        Jac = ['#Non-zero Elements',]
        jac_elements = chemicalsys.jacobian()
        for jac_ij in jac_elements: 
            var,expr = self.jacobian_element(jac_ij)
            if expr and expr != '0':
                Jac.append(f'{var} = {expr}')
        return Jac
    def _function(self,chemicalsys,level=0):
        """
        Generates the code of function 'f' in  dxdt = f(x,t) where x, dxdt and t
        are vectors. This function corresponds to the system of diferential
        equations for the mass balances of the system.
        """
        definition = f'def {self.f}(t,{self.x}):'

        lines = [ f'{self.dxdt} = np.zeros({chemicalsys.species})',]
        # Write the constants block
        lines.append('')
        constants = self._kinetic_constants(chemicalsys)
        lines.extend(constants)
        # Write the ratelaws block
        lines.append('')
        ratelaws = self._ratelaws(chemicalsys)
        lines.extend(ratelaws)
        # Write the MassBalances block
        lines.append('')
        massbalances = self._massbalances(chemicalsys)
        lines.extend(massbalances)
        # Function end
        lines.append('')
        lines.append(f'return {self.dxdt}')
        lines.append('\n')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add function definition
        lines.insert(0,definition)

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _jacobian(self,chemicalsys,level=0):
        """
        Generates the code of function 'Jac' in  dxdt = Jac(x,t) where x, dxdt
        and t are vectors. This function corresponds to the Jacobian of the
        system of differential equations for the mass balances of the system.
        """
        definition = f'def {self.jac_f}(t,{self.x}):'
        n = chemicalsys.species
        lines = [ f'{self.jac} = np.zeros(shape=({n},{n}))',]
        # Write the constants block
        lines.append('')
        constants = self._kinetic_constants(chemicalsys)
        lines.extend(constants)
        # Write the jacobian elements
        lines.append('')
        elements = self._jacobian_elements(chemicalsys)
        lines.extend(elements)
        # Function end
        lines.append('')
        lines.append(f'return {self.jac}')
        lines.append('\n')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add function definition
        lines.insert(0,definition)

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _initial_concentrations(self,simulation):
        concentrations = []
        for key,val in simulation['concentrations'].items():
            concentrations.append(f'xini[{key}] = {val}')
        return '\n'.join(Indent(concentrations,level=0))

    # main writing methods
    def fill_header(self,chemicalsys):
        out_filename = self.parameters.get('out_filename','data.txt')
        self.parameters['out_filename'] = out_filename
        self.parameters['species'] = chemicalsys.species
        self.parameters['T'] = chemicalsys.T
        super().fill_header(chemicalsys)
    def write(self,chemicalsys,filepath):
        self.fill(chemicalsys)
        # Write the constants block
        #lines.append('#Constants')
        items = [self.header,
                 self.function,
                 self.jacobian,
                 self.tail]
        with open(filepath,'w') as F:
            F.write(''.join(items))

class SemiBatch(Batch):
    """
    A writer to account for a SemiBatch Reactor with an influx of species 
    coupled after a 'tsection' time of the total simulation with a Batch reactor. 
    As a consequence the Simulation parameters required for this type of model 
    require the parameters 'tsection' and 'tfin' as well as the parameters for 
    the initial SemiBatch reactor.

    Parameters
    ----------
    flow : float
        Flowrate in L/s
    volume : float
        Reactor Volume in L
    """
    def __init__(self,flow,Vini,**kwargs):
        super().__init__(**kwargs)
        self.keys = [] # Keys of the different reactants added
        self.flow = flow
        self.Vini = Vini # initial volume
        self.Cadd = None
    def _load_default_header(self):
        with open(TEMPLATES_PATH / 'python_semibatch.head','r') as F:
            txt = F.read()
        self._header = txt
    def _load_default_tail(self):
        with open(TEMPLATES_PATH / 'python_semibatch.tail','r') as F:
            txt = F.read()
        self._tail = txt
    def set_parameters(self,**kwargs):
        super().set_parameters(**kwargs)

    # methods for object -> str transformations
    def massbalance(self,MassBalance):
        if MassBalance.compound.key not in self.keys:
            return super().massbalance(MassBalance)
        
        var, expr = super().massbalance(MassBalance)
        expr = f'tflow*Cin{MassBalance.compound.key:02.0f} {expr}'
        return var, expr

    # main writing methods
    def fill_header(self,chemicalsys):
        out_filename = self.parameters.get('out_filename','data.txt')
        self.parameters['out_filename'] = out_filename
        self.parameters['species'] = chemicalsys.species
        self.parameters['T'] = chemicalsys.T
        self.parameters['Vini'] = self.Vini
        self.parameters['Cadd'] = self.Cadd
        self.parameters['flow'] = self.flow
        super().fill_header(chemicalsys)

    # auxiliary writing methods
    def _function(self,chemicalsys,level=0):
        """
        Generates the code of function 'f' in  dxdt = f(x,t) where x, dxdt and t
        are vectors. This function corresponds to the system of diferential
        equations for the mass balances of the system.
        """
        definition = f'def {self.f}(t,{self.x}):'

        lines = [ f'{self.dxdt} = np.zeros({chemicalsys.species}+1)',]
        # Write the Inlet Concentrations and Flow
        lines.append(f'tflow = {self.flow}/{self.x}[-1] # flow/V; 1/s')
        inletC = self._flask_initial_concentrations()
        lines.extend(inletC)
        # Write the constants block
        lines.append('')
        constants = self._kinetic_constants(chemicalsys)
        lines.extend(constants)
        # Write the ratelaws block
        lines.append('')
        ratelaws = self._ratelaws(chemicalsys)
        lines.extend(ratelaws)
        # Write the MassBalances block
        lines.append('')
        massbalances = self._massbalances(chemicalsys)
        lines.extend(massbalances)
        # Include the Volume 
        lines.append('')
        lines.append(f'{self.dxdt}[-1] = tflow # flow/V; 1/s')
        lines.append(f'{self.dxdt}[:] = {self.dxdt}[:]*{self.x}[-1]')
        # Function end
        lines.append('')
        lines.append(f'return {self.dxdt}')
        lines.append('\n')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add function definition
        lines.insert(0,definition)

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _jacobian(self,chemicalsys,level=0):
        """
        Generates the code of function 'Jac' in  dxdt = Jac(x,t) where x, dxdt
        and t are vectors. This function corresponds to the Jacobian of the
        system of differential equations for the mass balances of the system.
        """
        definition = f'def {self.jac_f}(t,{self.x}):'
        n = chemicalsys.species
        lines = [ f'{self.jac} = np.zeros(shape=({n}+1,{n}+1))',]
        # Write the constants block
        lines.append('')
        constants = self._kinetic_constants(chemicalsys)
        lines.extend(constants)
        # Write the jacobian elements
        lines.append('')
        elements = self._jacobian_elements(chemicalsys)
        lines.extend(elements)
        # Include the Volume 
        lines.append('')
        lines.append(f'{self.jac}[:,:] = {self.jac}[:,:]*{self.x}[-1]')

        # Function end
        lines.append('')
        lines.append(f'return {self.jac}')
        lines.append('\n')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add function definition
        lines.insert(0,definition)

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _set_concentrations(self,concentrations):
        reactor = dict()
        flask = dict()
        for key,(r,f) in concentrations.items(): 
            if r != 0: 
                reactor[key] = r
            if f != 0: 
                flask[key] = f
        self.parameters['concentrations'] = reactor
        self.Cadd = flask
        self.keys = list(flask.keys())
    def _initial_concentrations(self,simulation):
        self._set_concentrations(simulation['concentrations'])
        return super()._initial_concentrations(self.parameters)
    def _flask_initial_concentrations(self): 
        concentrations = []
        for key,val in self.Cadd.items():
            concentrations.append(f'Cin{key:02.0f} = {val}')
        return concentrations
class SemiBatchExtended(SemiBatch):
    """
    Extension of the SemiBatch reactor model that allows up to 2 chemical 
    systems, one per section.
    """
    def _load_default_header(self):
        with open(TEMPLATES_PATH / 'python_semibatch_extended.head','r') as F:
            txt = F.read()
        self._header = txt
    def _load_default_tail(self):
        with open(TEMPLATES_PATH / 'python_semibatch_extended.tail','r') as F:
            txt = F.read()
        self._tail = txt
    def set_parameters(self,simulation=None,convergence=None):
        if simulation is not None:
            self.parameters.update(simulation)
            concentrations = self._initial_concentrations(simulation)
            self.parameters['concentrations'] = concentrations
        if convergence is not None:
            self.parameters['convergence'] = convergence.as_str(sep=',')

    def _function(self,systems,level=0):
        """
        Generates the code of function 'f' in  dxdt = f(x,t) where x, dxdt and t
        are vectors. This function corresponds to the system of diferential
        equations for the mass balances of the system.
        """
        _f = self.f
        self.f = f'{self.f}_semi'

        fadd = super()._function(systems[0],level)
        
        self.f = f'{_f}_batch'
        # Remove the "added compounds"
        _keys = self.keys
        self.keys = []

        fnoadd = super(SemiBatch,self)._function(systems[1],level)
        
        # Return everything to normal
        self.f = _f
        self.keys = _keys


        return fadd + fnoadd
    def _jacobian(self, systems, level=0):
        _jac_f = self.jac_f 
        self.jac_f = f'{self.jac_f}_semi'

        jacadd = super()._jacobian(systems[0],level)
        
        self.jac_f = f'{_jac_f}_batch'
        # Remove the "added compounds"
        _keys = self.keys
        self.keys = []

        jacnoadd = super(SemiBatch,self)._jacobian(systems[1],level)

        self.jac_f = _jac_f
        self.keys = _keys

        return jacadd + jacnoadd
    
    # main methods for writing
    def fill(self,system1,system2=None):
        """
        Reads the information of the chemical system and updates the values
        needed for writing.

        Parameters
        ----------
        system1 : ChemicalSystem
            Chemical System to be written for the SemiBatch reactor model section
        system2 : ChemicalSystem, optional
            Chemical System to be written for the Batch reactor model section, 
            If none is provided it defaults to the system1.
        """
        if system2 is None: 
            systems = [system1,system1]
        else: 
            systems = [system1,system2]
        self.fill_header(systems[0])
        self.fill_tail(systems[0])
        self.function = self._function(systems)
        self.jacobian = self._jacobian(systems)
class PFR(Batch):
    """
    Writer class of python scripts for ideal Plug Flow Reactors (PFR) models.
    Models using this writer take inlet concentrations as floats and not tuples.
    Allows providing the parameters through a SimulationParameters instance or 
    directly upon instantiation.
    
    Parameters
    ----------
    flow : float
        Flowrate in L/s
    volume : float
        Reactor Volume in L

    """
    def __init__(self,flow=1.0,volume=1.0,**kwargs):
        super().__init__(**kwargs)
        self.flow = flow # L/s
        self.volume = volume #L
    def _load_default_header(self):
        with open(TEMPLATES_PATH / 'python_PFR.head','r') as F:
            txt = F.read()
        self._header = txt
    def _load_default_tail(self):
        with open(TEMPLATES_PATH / 'python_PFR.tail','r') as F:
            txt = F.read()
        self._tail = txt
    def set_parameters(self,**kwargs):
        super().set_parameters(**kwargs)
        self.parameters['flow'] = self.parameters.get('flow',self.flow)
        self.parameters['volume'] = self.parameters.get('volume',self.volume)

    # main writing methods
    def fill_header(self,chemicalsys):
        out_filename = self.parameters.get('out_filename','data.txt')
        self.parameters['out_filename'] = out_filename
        self.parameters['species'] = chemicalsys.species
        self.parameters['T'] = chemicalsys.T
        super().fill_header(chemicalsys)

    def _function(self,chemicalsys,level=0):
        """
        Generates the code of function 'f' in  dxdt = f(x,t) where x, dxdt and t
        are vectors. This function corresponds to the system of diferential
        equations for the mass balances of the system.
        """
        definition = f'def {self.f}(t,{self.x}):'

        lines = [ f'{self.dxdt} = np.zeros({chemicalsys.species})',]
        # Write the Inlet Concentrations and Flow
        lines.append(f'flow = {self.flow} # flow; L/s')
        # Write the constants block
        lines.append('')
        constants = self._kinetic_constants(chemicalsys)
        lines.extend(constants)
        # Write the ratelaws block
        lines.append('')
        ratelaws = self._ratelaws(chemicalsys)
        lines.extend(ratelaws)
        # Write the MassBalances block
        lines.append('')
        massbalances = self._massbalances(chemicalsys)
        lines.extend(massbalances)
        # Include the Volume 
        lines.append('')
        lines.append(f'{self.dxdt}[:] = {self.dxdt}[:]/flow')
        # Function end
        lines.append('')
        lines.append(f'return {self.dxdt}')
        lines.append('\n')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add function definition
        lines.insert(0,definition)

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _jacobian(self,chemicalsys,level=0):
        """
        Generates the code of function 'Jac' in  dxdt = Jac(x,t) where x, dxdt
        and t are vectors. This function corresponds to the Jacobian of the
        system of differential equations for the mass balances of the system.
        """
        definition = f'def {self.jac_f}(t,{self.x}):'
        n = chemicalsys.species
        lines = [ f'{self.jac} = np.zeros(shape=({n},{n}))',]
        lines.append(f'flow = {self.flow} # flow; L/s')
        # Write the constants block
        lines.append('')
        constants = self._kinetic_constants(chemicalsys)
        lines.extend(constants)
        # Write the jacobian elements
        lines.append('')
        elements = self._jacobian_elements(chemicalsys)
        lines.extend(elements)
        # Include the Volume 
        lines.append('')
        lines.append(f'{self.jac}[:,:] = {self.jac}[:,:]/flow')

        # Function end
        lines.append('')
        lines.append(f'return {self.jac}')
        lines.append('\n')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add function definition
        lines.insert(0,definition)

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    



"""
This module contains all the different writers per reactor for the C++ language.
"""

from collections import Counter
from ._base import TEMPLATES_PATH, Indent, Writer

class CplusplusWriter(Writer):

    def _load_default_header(self):
        with open(TEMPLATES_PATH / 'cplusplus_header.default','r') as F:
            txt = F.read()
        self._header = txt
    def _load_default_tail(self):
        with open(TEMPLATES_PATH / 'cplusplus_tail.default','r') as F:
            txt = F.read()
        self._tail = txt

    def set_parameters(self,simulation=None,convergence=None):
        if simulation is not None:
            for key in ['tfin','trep','dt']: 
                self.parameters[key] = simulation.get(key,'')
            concentrations = self._initial_concentrations(simulation)
            self.parameters['concentrations'] = concentrations
        if convergence is not None:
            variables = ','.join(convergence.variables)
            values = convergence.as_str(sep=';')
            lines = [f'double {variables};',
                     f"    {values};"]
            self.parameters['convergence'] = '\n'.join(Indent(lines,level=0))

    # methods for object -> str transformations
    def ratelaw(self,reaction): # Currently only for Elemental Steps
        var = f'r{reaction.key:02.0f}'
        Aux = []
        expr = f'k{reaction.key:02.0f}*'
        reactants = reaction.reactants.elements() # Unpack the reactants Counter
        for R in sorted(reactants,key=lambda x: x.key):
            Aux.append(f'{self.x}[{R.key}]')
        expr += '*'.join(Aux)
        return  var, expr
    def ratelaw_partial(self,reaction,compound): # Currently only for Elemental Steps
        #var = f'r{reaction.key:02.0f}' Legacy code, to remove
        if compound not in reaction.reactants:
            return ''
        Aux = []
        # Create a new Unpack the reactants Counter
        reactants = Counter(reaction.reactants.elements())
        coef = reactants[compound]
        if coef > 1:
            Aux.append(f'{coef:0.1f}*k{reaction.key:02.0f}')
        else:
            Aux.append(f'k{reaction.key:02.0f}')
        reactants[compound] -= 1
        reactants = reactants.elements()
        for R in sorted(reactants,key=lambda x: x.key):
            Aux.append(f'{self.x}[{R.key}]')
        expr = '*'.join(Aux)
        return  expr
    def constant(self,reaction,value_format='{:0.10e}'):  # Currently only for Elemental Steps
        return super().constant(reaction,value_format)
    def massbalance(self,MassBalance):
        if MassBalance.compound.key is None:
            return '', '0'
        var = f'{self.dxdt}[{MassBalance.compound.key}]'
        # Assume that if it has a key, it is a compound of the Chemical System
        if not MassBalance.items: 
            return var, '0'
        expr = []
        for coef,reaction in MassBalance.items:
            if coef == 1:
                Aux = '+'
            elif coef == -1:
                Aux = '-'
            else:
                Aux = f'{coef:+0.1f}*'
            r_expr = f'{Aux}r{reaction.key:02.0f}'
            expr.append(r_expr)
        return var, ''.join(expr)
    def jacobian_element(self,Jac_ij):
        C1,C2 = Jac_ij.compound1, Jac_ij.compound2
        var = f'{self.jac}({C1.key},{C2.key})'
        if C1.key is None or C2.key is None: 
            return '', '0'
        expr = []
        for coef,reaction in Jac_ij.items:
            if coef == 1:
                Aux = '+'
            elif coef == -1:
                Aux = '-'
            else:
                Aux = f'{coef:+0.1f}*'
            r_expr = self.ratelaw_partial(reaction,C2)
            if not r_expr:
                continue
            expr.append(Aux + r_expr)
        if not expr:
            return var ,'0'
        return var, ''.join(expr)

    # auxiliary writing methods
    def _kinetic_constants(self,chemicalsys):
        constants = ['// Kinetic Constants',]
        for reaction in chemicalsys.reactions:
            # vars,values = # In case we consider supporting reversible elemental steps?
            var,value = self.constant(reaction)
            constants.append(f'const double {var} = {value};')
        return constants
    def _ratelaws(self,chemicalsys):
        vars = []
        ratelaws = ['// Ratelaws',]
        for reaction in chemicalsys.reactions:
            var,law = self.ratelaw(reaction)
            if var:
                ratelaws.append(f'{var} = {law};')
            vars.append(var)
        # Add variable declaration
        ratelaws.insert(1,f'double {",".join(vars)};')
        return ratelaws
    def _massbalances(self,chemicalsys):
        MBs = ['// MassBalances',]
        for compound in chemicalsys.compounds:
            MB = chemicalsys.massbalance(compound)
            var,law = self.massbalance(MB)
            if var:
                MBs.append(f'{var} = {law};')
        return MBs
    def _jacobian_elements(self,chemicalsys):
        Jac = []
        jac_elements = chemicalsys.jacobian()
        for jac_ij in jac_elements: 
            var,expr = self.jacobian_element(jac_ij)
            if expr:
                Jac.append(f'{var} = {expr};')
        return Jac
    def _function(self,chemicalsys,level=0):
        """
        Generates the code of function 'f' in  dxdt = f(x,t) where x, dxdt and t
        are vectors. This function corresponds to the system of diferential
        equations for the mass balances of the system.
        """
        definition = f'void operator()(const state_type &{self.x}, '\
                     f'state_type &{self.dxdt}, const double t)'

        lines = []

        # Write the ratelaws block
        ratelaws = self._ratelaws(chemicalsys)
        lines.extend(ratelaws)

        # Write the MassBalances block
        massbalances = self._massbalances(chemicalsys)
        lines.extend(massbalances)
        # Function end

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add brackets
        lines.insert(0,'{')
        lines.append('}')
        # Add function definition
        lines.insert(0,definition)

        # indent once
        lines = Indent(lines,tab='    ',level=1)

        # Add brackets
        lines.insert(0,'{')
        lines.append('};')
        # Add structure name
        lines.insert(0,f'struct {self.f}')

        # Add the indentation level of the function
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _jacobian(self,chemicalsys,level=0):
        """
        Generates the code of the jacobian of the function 'f' in dxdt = f(x,t) 
        where x, dxdt and t are vectors.
        """
        definition = f'void operator()(const state_type &{self.x}, '\
                     f'matrix_type &{self.jac}, const double t, '\
                     f'state_type &dfdt)'

        lines = []

        # Write the jacobian elements
        elements = self._jacobian_elements(chemicalsys)
        lines.extend(elements)
        # Function end
        for i in range(chemicalsys.species): 
            lines.append(f'dfdt[{i}] = 0;')

        # Add one indentation level
        lines = Indent(lines,tab='    ',level=1)

        # Add brackets
        lines.insert(0,'{')
        lines.append('}')
        # Add function definition
        lines.insert(0,definition)

        # indent once
        lines = Indent(lines,tab='    ',level=1)

        # Add brackets
        lines.insert(0,'{')
        lines.append('};')
        # Add structure name
        lines.insert(0,f'struct {self.jac_f}')

        # Ensure the final indentation level
        lines = Indent(lines,tab='    ',level=level)

        return '\n'.join(lines)
    def _initial_concentrations(self,simulation):
        concentrations = []
        for key,val in simulation['concentrations'].items():
            concentrations.append(f'{self.x}[{key}] = {val};')
        concentrations[1:] = Indent(concentrations[1:],tab='    ',level=1)
        return '\n'.join(concentrations)

    # main writing methods
    def fill_header(self,chemicalsys):
        self.header = self._header.format_map(self.parameters)
    def fill_tail(self,chemicalsys):
        self.parameters['species'] = chemicalsys.species
        self.parameters['T'] = chemicalsys.T
        super().fill_tail(chemicalsys)
    def fill(self,chemicalsys):
        super().fill(chemicalsys)
        constants = self._kinetic_constants(chemicalsys)
        self.constants = '\n'.join(constants)
    def clear(self):
        super().clear()
        self.constants = ''

    def write(self,chemicalsys,filepath):
        self.fill(chemicalsys)
        items = [self.header,
                 self.constants,
                 '',
                 self.function,
                 '',
                 self.jacobian,
                 '',
                 self.tail]
        with open(filepath,'w') as F:
            F.write('\n'.join(items))
